#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Wrapper around the video content type.
"""

import typing as t

from threatexchange.content_type.content_base import ContentType


class VideoContent(ContentType):
    """
    Content representing a sequence of images, giving the illusion of motion.

    Examples might be:
    * mp4
    * avi
    * gif animations
    """



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Common interface for persisting pytx configuration and concepts.

Most of the individual components of pytx are find to use piecemeal, and
the full interface covers the most complex and complete useage. A usecase
with one collection of hashes using one algorithm might be better off
hardcoding those things rather than fully implementing the interface.

# Migration Notes
There's an earlier attempt at these interfaces used for CLI at
<@Mackay-Fisher add the right pointer to the CLI storage>.
During the development of Hasher-Matcher-Actioner 2.0
(github.com/facebook/ThreatExchange/tree/main/hasher-matcher-actioner/)
we realized that the original attempt at this wouldn't meet the needs
of that code and wrote a new interface.

As of 12/2024, we are now migrating that interface from HMA into pytx
proper as part of a migration to dbm
(github.com/facebook/ThreatExchange/issues/1687). The general approach is:
1. Copy the interface unchanged from HMA to pytx
2. Release a new version of pytx
3. Delete the copy in HMA and update all references to the pytx version

In parallel, we intend to slowly migrate the CLI storage components to
follow the same interface.
"""

import abc
from dataclasses import dataclass
import typing as t
from threatexchange.content_type.content_base import ContentType
from threatexchange.signal_type.signal_base import SignalType


@dataclass
class SignalTypeConfig:
    """
    Holder for SignalType configuration
    """

    # Signal types that are not enabled should not be used in hashing/matching
    enabled_ratio: float
    signal_type: t.Type[SignalType]

    @property
    def enabled(self) -> bool:
        # TODO do a coin flip here, but also refactor this to do seeding
        return self.enabled_ratio >= 0.0


class ISignalTypeConfigStore(metaclass=abc.ABCMeta):
    """Interface for accessing SignalType configuration"""

    @abc.abstractmethod
    def get_signal_type_configs(self) -> t.Mapping[str, SignalTypeConfig]:
        """Return all installed signal types."""

    @abc.abstractmethod
    def _create_or_update_signal_type_override(
        self, signal_type: str, enabled_ratio: float
    ) -> None:
        """Create or update database entry for a signal type, setting a new value."""

    @t.final
    def create_or_update_signal_type_override(
        self, signal_type: str, enabled_ratio: float
    ) -> None:
        """Update enabled ratio of an installed signal type."""
        installed_signal_types = self.get_signal_type_configs()
        if signal_type not in installed_signal_types:
            raise ValueError(f"Unknown signal type {signal_type}")
        if not (0.0 <= enabled_ratio <= 1.0):
            raise ValueError(
                f"Invalid enabled ratio {enabled_ratio}. Must be in the range 0.0-1.0 inclusive."
            )
        self._create_or_update_signal_type_override(signal_type, enabled_ratio)

    @t.final
    def get_enabled_signal_types(self) -> t.Mapping[str, t.Type[SignalType]]:
        """Helper shortcut for getting only enabled SignalTypes"""
        return {
            k: v.signal_type
            for k, v in self.get_signal_type_configs().items()
            if v.enabled
        }

    @t.final
    def get_enabled_signal_types_for_content_type(
        self, content_type: t.Type[ContentType]
    ) -> t.Mapping[str, t.Type[SignalType]]:
        """Helper shortcut for getting enabled types for a piece of content"""
        return {
            k: v.signal_type
            for k, v in self.get_signal_type_configs().items()
            if v.enabled and content_type in v.signal_type.get_content_types()
        }


@dataclass
class ContentTypeConfig:
    """
    Holder for ContentType configuration.
    """

    # Content types that are not enabled should not be used in hashing/matching
    enabled: bool
    content_type: t.Type[ContentType]


class IContentTypeConfigStore(metaclass=abc.ABCMeta):
    """Interface for accessing ContentType configuration"""

    @abc.abstractmethod
    def get_content_type_configs(self) -> t.Mapping[str, ContentTypeConfig]:
        """
        Return all installed content types.
        """



# Copyright (c) Meta Platforms, Inc. and affiliates.

from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.extensions.pdf.content import PDFContent

TX_MANIFEST = ThreatExchangeExtensionManifest(
    content_types=(PDFContent,),
)



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Wrapper around the Photo PDQ signal type.
"""

import typing as t
import re
import random

from threatexchange.signal_type.pdq.pdq_hasher import pdq_from_bytes
from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.photo import PhotoContent
from threatexchange.signal_type import signal_base
from threatexchange.signal_type.pdq.pdq_utils import (
    simple_distance,
    PDQ_CONFIDENT_MATCH_THRESHOLD,
)
from threatexchange.exchanges.impl.fb_threatexchange_signal import (
    HasFbThreatExchangeIndicatorType,
)
from threatexchange.signal_type.pdq.pdq_index import PDQIndex


class PdqSignal(
    signal_base.SimpleSignalType,
    signal_base.BytesHasher,
    HasFbThreatExchangeIndicatorType,
    signal_base.CanGenerateRandomSignal,
):
    """
    PDQ is an open source photo similarity algorithm.

    Unlike MD5s, which are sensitive to single pixel differences, PDQ has
    a concept of "distance" and can detect when content is visually similar.
    This property tends to make it much more effective at finding images that
    a human would claim are the same, but also opens the door for false
    positives.

    Which distance to use can differ based on the type of content being
    searched for. While the PDQ documentation suggests certain thresholds,
    they can sometimes vary depending on what you are comparing against.
    """

    INDICATOR_TYPE = "HASH_PDQ"

    # This may need to be updated (TODO make more configurable)
    # Images with less than quality 50 are too unreliable to match on
    QUALITY_THRESHOLD = 50

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [PhotoContent]

    @classmethod
    def get_index_cls(cls) -> t.Type[PDQIndex]:
        return PDQIndex

    @classmethod
    def validate_signal_str(cls, signal_str: str) -> str:
        """PDQ hash contains 64 hexidecimal characters."""
        if not re.match("^[0-9a-f]{64}$", signal_str):
            raise ValueError("invalid PDQ hash")
        return signal_str

    @classmethod
    def compare_hash(
        cls,
        hash1: str,
        hash2: str,
        pdq_dist_threshold: int = PDQ_CONFIDENT_MATCH_THRESHOLD,
    ) -> signal_base.SignalComparisonResult:
        dist = simple_distance(hash1, hash2)
        return signal_base.SignalComparisonResult.from_simple_dist(
            dist, pdq_dist_threshold
        )

    @classmethod
    def hash_from_bytes(cls, bytes_: bytes) -> str:
        pdq_hash, quality = pdq_from_bytes(bytes_)
        if quality < cls.QUALITY_THRESHOLD:
            return ""
        return pdq_hash

    @classmethod
    def get_random_signal(cls) -> str:
        # Generate a random hexadecimal string of length 64
        return f"{random.randrange(16**64):064x}"

    @staticmethod
    def get_examples() -> t.List[str]:
        return [
            # pdq/data/bridge-mods
            "f8f8f0cee0f4a84f06370a22038f63f0b36e2ed596621e1d33e6b39c4e9c9b22",
            "f8f8f0cee0f4a84f06370a2a038f63f0b36e26d596621e1d33e6b39c4e9c9b22",
            "f8f8f0cee0f4a84f0637022a038f67f0b36e26d596621e1d33e6b39c4e9c9b22",
            "f8f8f0cee0f4a84f06370a2a068f67f0b36e26d596621e1d33e6339c4e9c9b22",
            "f8f8f0cee0f4a84f06370a22038f67f0b36e2ed596221e1d33e6b39c4e9c9b22",
            "f8f8f0cee0f4a84f0e370a22038f67f0b36e2ed596221e1d33e6b39c4a9c9b22",
            "d8f8f0cec0f4a84f8e370a22038f67f0b36a2ed597231e1d72e6b39c4a9c9b22",
            "f8f8f0cee0f4a84f06370a22038f67f0b36e2ed596621e1d33e6339c4e9c9b22",
            "d0f8f1ccc0f4a84d0a370a3a228f67f0b36e2ed5b6623e1d33e6339c4e9c9b22",
            "d8f8f1eec0f4a84f0e37022a078f63f0b36e2ed596621e1d33e6239c4e9c9b22",
            "d8f8f0cec4f4a84f0637022a078f67f0b36e2ee5b6621e1d33e6239c4e9c9b22",
            "d8f8f0cec0f4a84f0637022a278f67f0b36e2ed596621e1d33e6339c4e9c9b22",
            # pdq/data/misc-images
            "e64cc9d91c623882f8d1f1d9a398e78c9f199b3bd83924f2b7e11e0bf861b064",
            "6227401f601ff4ccafcc9fad4b0d95d371a2eb7265a3285234d228ca94deeb2d",
            # Github invertocat - https://github.com/logos
            "c186f9619659658c4b6916daf934496985f4259c72c39676ab9edb631c3c349c",
            # Github octocat - https://github.com/logos
            "62c9789627838f69f23cf98c29e37836d61c87c31c59587c4b49a783f496725c",
            # https://github.com/facebook/ThreatExchange/issues/1609 - bad entry
            # "00000000000000000000000000000000",
        ]



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Implementation of SignalTypeIndex abstraction for PDQ
"""

import typing as t
import faiss
import numpy as np
import weakref


from threatexchange.signal_type.index import (
    IndexMatchUntyped,
    SignalSimilarityInfoWithIntDistance,
    SignalTypeIndex,
    T as IndexT,
)
from threatexchange.signal_type.pdq.pdq_utils import (
    BITS_IN_PDQ,
    PDQ_CONFIDENT_MATCH_THRESHOLD,
    convert_pdq_strings_to_ndarray,
)

PDQIndexMatch = IndexMatchUntyped[SignalSimilarityInfoWithIntDistance, IndexT]


class PDQIndex2(SignalTypeIndex[IndexT]):
    """
    Indexing and querying PDQ signals using Faiss for approximate nearest neighbor search.

    This is a redo of the existing PDQ index,
    designed to be simpler and fix hard-to-squash bugs in the existing implementation.
    Purpose of this class: to replace the original index in pytx 2.0
    """

    def __init__(
        self,
        index: t.Optional[faiss.Index] = None,
        entries: t.Iterable[t.Tuple[str, IndexT]] = (),
        *,
        threshold: int = PDQ_CONFIDENT_MATCH_THRESHOLD,
    ) -> None:
        super().__init__()
        self.threshold = threshold

        if index is None:
            index = faiss.IndexFlatL2(BITS_IN_PDQ)
        self._index = _PDQFaissIndex(index)

        # Matches hash to Faiss index
        self._deduper: t.Dict[str, int] = {}
        # Entry mapping: Each list[entries]'s index is its hash's index
        self._idx_to_entries: t.List[t.List[IndexT]] = []

        self.add_all(entries=entries)

    def __len__(self) -> int:
        return len(self._idx_to_entries)

    def query(self, hash: str) -> t.Sequence[PDQIndexMatch[IndexT]]:
        """
        Look up entries against the index, up to the threshold.
        """
        results: t.List[PDQIndexMatch[IndexT]] = []
        matches_list: t.List[t.Tuple[int, int]] = self._index.search(
            queries=[hash], threshold=self.threshold
        )

        for match, distance in matches_list:
            entries = self._idx_to_entries[match]
            # Create match objects for each entry
            results.extend(
                PDQIndexMatch(
                    SignalSimilarityInfoWithIntDistance(distance=distance),
                    entry,
                )
                for entry in entries
            )
        return results

    def add(self, signal_str: str, entry: IndexT) -> None:
        self.add_all(((signal_str, entry),))

    def add_all(self, entries: t.Iterable[t.Tuple[str, IndexT]]) -> None:
        for h, i in entries:
            existing_faiss_id = self._deduper.get(h)
            if existing_faiss_id is None:
                self._index.add([h])
                self._idx_to_entries.append([i])
                next_id = len(self._deduper)  # Because faiss index starts from 0 up
                self._deduper[h] = next_id
            else:
                # Since this already exists, we don't add it to Faiss because Faiss cannot handle duplication
                self._idx_to_entries[existing_faiss_id].append(i)


class _PDQFaissIndex:
    """
    A wrapper around the faiss index for pickle serialization
    """

    def __init__(self, faiss_index: faiss.Index) -> None:
        self.faiss_index = faiss_index
        self._finalizer = weakref.finalize(
            self, _PDQFaissIndex._finalize_faiss, self.faiss_index
        )

    def add(self, pdq_strings: t.Sequence[str]) -> None:
        """
        Add PDQ hashes to the FAISS index.
        """
        vectors = convert_pdq_strings_to_ndarray(pdq_strings)
        self.faiss_index.add(vectors)

    def search(
        self, queries: t.Sequence[str], threshold: int
    ) -> t.List[t.Tuple[int, int]]:
        """
        Search the FAISS index for matches to the given PDQ queries.
        """
        query_array: np.ndarray = convert_pdq_strings_to_ndarray(queries)
        limits, distances, indices = self.faiss_index.range_search(
            query_array, threshold + 1
        )

        results: t.List[t.Tuple[int, int]] = []
        for i in range(len(queries)):
            matches = [idx.item() for idx in indices[limits[i] : limits[i + 1]]]
            dists = [dist for dist in distances[limits[i] : limits[i + 1]]]
            for j in range(len(matches)):
                results.append((matches[j], dists[j]))
        return results

    def __getstate__(self):
        return faiss.serialize_index(self.faiss_index)

    def __setstate__(self, data):
        self.faiss_index = faiss.deserialize_index(data)
        self._finalizer = weakref.finalize(
            self, _PDQFaissIndex._finalize_faiss, self.faiss_index
        )

    def dispose(self) -> None:
        try:
            reset_fn = getattr(self.faiss_index, "reset", None)
            if callable(reset_fn):
                reset_fn()
        except Exception:
            pass

    @staticmethod
    def _finalize_faiss(faiss_index: faiss.Index) -> None:
        try:
            reset_fn = getattr(faiss_index, "reset", None)
            if callable(reset_fn):
                reset_fn()
        except Exception:
            pass



# Copyright (c) Meta Platforms, Inc. and affiliates.

from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.extensions.pdq_ocr.pdq_ocr import PdqOcrSignal

TX_MANIFEST = ThreatExchangeExtensionManifest(
    signal_types=(PdqOcrSignal,),
)



# Copyright (c) Meta Platforms, Inc. and affiliates.

import typing as t
from dataclasses import dataclass, field

from threatexchange.exchanges import auth, fetch_state as state, signal_exchange_api
from threatexchange.exchanges.clients.techagainstterrorism import api
from threatexchange.exchanges.collab_config import CollaborationConfigWithDefaults
from threatexchange.signal_type.md5 import VideoMD5Signal
from threatexchange.signal_type.pdq.signal import PdqSignal
from threatexchange.signal_type.signal_base import SignalType


@dataclass
class TATCheckpoint(state.FetchCheckpointBase):
    checkpoint: str

    def get_progress_timestamp(self) -> t.Optional[int]:
        if self.checkpoint:
            return int(float(self.checkpoint.partition(",")[0]))
        return None

    @classmethod
    def from_tat_fetch(cls, response: api.TATHashListResponse) -> "TATCheckpoint":
        return cls(response.checkpoint)


@dataclass
class TATCredentials(auth.CredentialHelper):
    ENV_VARIABLE: t.ClassVar[str] = "PYTX_TAT_CREDENTIALS"
    FILE_NAME: t.ClassVar[str] = "~/.pytx_tat_credentials"

    username: str = field(
        metadata={"help": "Tech Against Terrorism hash list API username."}
    )
    password: str = field(
        metadata={"help": "Tech Against Terrorism hash list API password."}
    )

    @classmethod
    def _from_str(cls, s: str) -> "TATCredentials":
        user, _, passw = s.strip().partition(":")
        return cls(user, passw)

    def _are_valid(self) -> bool:
        return bool(self.username and self.password)


class TATSignalExchangeAPI(
    auth.SignalExchangeWithAuth[CollaborationConfigWithDefaults, TATCredentials],
    signal_exchange_api.SignalExchangeAPIWithSimpleUpdates[
        CollaborationConfigWithDefaults,
        TATCheckpoint,
        state.FetchedSignalMetadata,
    ],
):

    def __init__(self, username: str, password: str) -> None:
        super().__init__()
        self.username = username
        self.password = password

    @staticmethod
    def get_config_cls() -> t.Type[CollaborationConfigWithDefaults]:
        return CollaborationConfigWithDefaults

    @staticmethod
    def get_checkpoint_cls() -> t.Type[TATCheckpoint]:
        return TATCheckpoint

    @staticmethod
    def get_record_cls() -> t.Type[state.FetchedSignalMetadata]:
        return state.FetchedSignalMetadata

    @staticmethod
    def get_credential_cls() -> t.Type[TATCredentials]:
        return TATCredentials

    @classmethod
    def get_name(cls) -> str:
        return "tat"

    @classmethod
    def for_collab(
        cls,
        collab: CollaborationConfigWithDefaults,
        credentials: t.Optional["TATCredentials"] = None,
    ) -> "TATSignalExchangeAPI":
        credentials = credentials or TATCredentials.get(cls)
        return cls(username=credentials.username, password=credentials.password)

    def get_client(self) -> api.TATHashListAPI:
        return api.TATHashListAPI(username=self.username, password=self.password)

    def fetch_iter(
        self,
        _supported_signal_types: t.Sequence[t.Type[SignalType]],
        checkpoint: t.Optional[TATCheckpoint],
    ) -> t.Iterator[
        state.FetchDelta[t.Tuple[str, str], state.FetchedSignalMetadata, TATCheckpoint]
    ]:

        client = self.get_client()

        _checkpoint = ""

        if checkpoint is not None:
            _checkpoint = checkpoint.checkpoint

        for result in client.fetch_hashes_iter(_checkpoint):

            translated = (_get_delta_mapping(r) for r in result.results)
            yield state.FetchDelta(
                dict(t for t in translated if t[0][0]),
                TATCheckpoint(result.checkpoint) if result.checkpoint else checkpoint,  # type: ignore[arg-type]
            )


def _is_compatible_signal_type(record: api.TATHashListEntry) -> bool:
    compatible_video_types = ["mov", "m4v", "mp4", "webm"]
    algorithm = record.algorithm

    if algorithm == "MD5":
        return record.file_type in compatible_video_types

    return algorithm == "PDQ"


def _type_mapping() -> t.Dict[str, str]:
    return {
        "PDQ": PdqSignal.get_name(),
        "MD5": VideoMD5Signal.get_name(),
    }


def _get_delta_mapping(
    record: api.TATHashListEntry,
) -> t.Tuple[t.Tuple[str, str], t.Optional[state.FetchedSignalMetadata]]:

    if not _is_compatible_signal_type(record):
        return (("", ""), None)

    type_str = _type_mapping().get(record.algorithm, "")
    metadata = state.FetchedSignalMetadata()

    return ((type_str, record.hash_digest), None if record.deleted else metadata)



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Helpers for implementing SignalExchangeAPIs or state.

They demonstrate how the methods in SignalExchangeAPI can be used to
reproduce databases of signals, as well as simple recipes for stage.
"""

from dataclasses import dataclass, field
import logging
import typing as t
from threatexchange.exchanges.signal_exchange_api import TSignalExchangeAPICls

from threatexchange.signal_type.signal_base import SignalType
from threatexchange.exchanges import fetch_state
from threatexchange.exchanges.collab_config import CollaborationConfigBase

K = t.TypeVar("K")
V = t.TypeVar("V")


@dataclass
class SimpleFetchedSignalMetadata(fetch_state.FetchedSignalMetadata):
    """
    Simple dataclass for fetched data.

    Merge by addition rather than replacement.
    """

    opinions: t.List[fetch_state.SignalOpinion] = field(default_factory=list)

    def get_as_opinions(self) -> t.Sequence[fetch_state.SignalOpinion]:
        return self.opinions


@dataclass
class _StateTracker:
    api_cls: TSignalExchangeAPICls
    _delta: t.Optional[fetch_state.FetchDeltaTyped]
    dirty: bool = False

    @property
    def empty(self) -> bool:
        return self._delta is None

    @property
    def delta(self) -> fetch_state.FetchDeltaTyped:
        assert self._delta is not None
        return self._delta

    @delta.setter
    def delta(self, value: fetch_state.FetchDeltaTyped) -> None:
        if self._delta is None:
            old = {}
            self._delta = value
        else:
            old = self._delta.updates
        self.api_cls.naive_fetch_merge(old, value.updates)
        self._delta.updates = old
        self._delta.checkpoint = value.checkpoint
        self.dirty = True

    @property
    def checkpoint(self) -> t.Optional[fetch_state.FetchCheckpointBase]:
        return None if self._delta is None else self._delta.checkpoint


class SimpleFetchedStateStore(fetch_state.FetchedStateStoreBase):
    """
    A FetchedStateStore that does merges in memory and writes by collab
    """

    def __init__(
        self,
        api_cls: TSignalExchangeAPICls,
    ) -> None:
        self.api_cls = api_cls
        self._state: t.Dict[str, _StateTracker] = {}

    def _read_state(
        self,
        collab_name: str,
    ) -> t.Optional[fetch_state.FetchDeltaTyped]:
        raise NotImplementedError

    def _write_state(
        self, collab_name: str, delta: fetch_state.FetchDeltaTyped
    ) -> None:
        raise NotImplementedError

    def get_checkpoint(
        self, collab: CollaborationConfigBase
    ) -> t.Optional[fetch_state.FetchCheckpointBase]:
        return self._get_state(collab).checkpoint

    def _get_state(self, collab: CollaborationConfigBase) -> _StateTracker:
        if collab.name not in self._state:
            logging.debug("Loading state for %s", collab.name)
            delta = self._read_state(collab.name)
            self._state[collab.name] = _StateTracker(self.api_cls, delta)
        return self._state[collab.name]

    def clear(self, collab: CollaborationConfigBase) -> None:
        self._state.pop(collab.name, None)

    def merge(
        self,
        collab: CollaborationConfigBase,
        delta: fetch_state.FetchDeltaTyped,
    ) -> None:
        """
        Merge a FetchDeltaBase into the state.

        At the implementation's discretion, it may call flush() or the
        equivalent work.
        """

        state = self._get_state(collab)
        if len(delta.updates) == 0 and delta.checkpoint in (
            None,
            state.checkpoint,
        ):
            logging.warning("No op update for %s", collab.name)
            return
        state.delta = delta

    def flush(self) -> None:
        for collab_name, state in self._state.items():
            if state.dirty:
                assert state.delta is not None
                self._write_state(collab_name, state.delta)
                state.dirty = False

    def get_for_signal_type(
        self, collabs: t.List[CollaborationConfigBase], signal_type: t.Type[SignalType]
    ) -> t.Dict[str, t.Dict[str, fetch_state.FetchedSignalMetadata]]:
        ret = {}
        for collab in collabs:
            state = self._get_state(collab)
            if not state.empty:
                by_signal = state.api_cls.naive_convert_to_signal_type(
                    [signal_type], collab, state.delta.updates
                ).get(signal_type, {})
                if by_signal:
                    ret[collab.name] = by_signal
        return ret



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
A wrapper around loading and storing ThreatExchange data from files.

There are a few categories of state that this wraps:
  1. Checkpoints - state about previous fetches
  2. Collaboration Indicator Dumps - Raw output from threat_updates
  3. Index state - serializations of indexes for SignalType
"""

import pickle
import pathlib
import typing as t
import logging

from threatexchange.signal_type.index import SignalTypeIndex
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.cli.exceptions import CommandError
from threatexchange.exchanges.collab_config import CollaborationConfigBase
from threatexchange.exchanges.fetch_state import (
    FetchDelta,
    FetchDeltaTyped,
)
from threatexchange.exchanges import helpers
from threatexchange.exchanges.signal_exchange_api import (
    SignalExchangeAPI,
    SignalExchangeAPIWithSimpleUpdates,
)
from threatexchange.signal_type import signal_base
from threatexchange.signal_type import index


class CliIndexStore:
    """
    Persistance layer for SignalTypeIndex objects for the cli.

    They are just stored to a file directory, with names based on their type.
    """

    FILE_EXTENSION = ".index"

    def __init__(self, indice_dir: pathlib.Path) -> None:
        self.dir = indice_dir

    def list(self) -> t.List[str]:
        """Return the names (SignalType.get_name()) of stored indices"""
        return [
            str(f)[: -len(self.FILE_EXTENSION)]
            for f in self.dir.glob(f"*{self.FILE_EXTENSION}")
        ]

    def clear(
        self, only_types: t.Optional[t.Iterable[t.Type[SignalType]]] = None
    ) -> None:
        """Clear persisted indices"""
        only_names = None
        if only_types is not None:
            only_names = {st.get_name() for st in only_types}
        for file in self.dir.glob(f"*{self.FILE_EXTENSION}"):
            if (
                only_names is None
                or str(file)[: -len(self.FILE_EXTENSION)] in only_names
            ):
                logging.info("Removing index %s", file)
                file.unlink()

    def _index_file(self, signal_type: t.Type[signal_base.SignalType]) -> pathlib.Path:
        """The expected path for the index for a signal type"""
        return self.dir / f"{signal_type.get_name()}{self.FILE_EXTENSION}"

    def store(
        self, signal_type: t.Type[signal_base.SignalType], index: SignalTypeIndex
    ) -> None:
        """Persist a SignalTypeIndex to disk"""
        assert signal_type.get_index_cls() == index.__class__
        path = self._index_file(signal_type)
        with path.open("wb") as fout:
            index.serialize(fout)

    def load(
        self, signal_type: t.Type[signal_base.SignalType]
    ) -> t.Optional[index.SignalTypeIndex]:
        """Load the SignalTypeIndex for this type from disk"""
        path = self._index_file(signal_type)
        if not path.exists():
            return None
        with path.open("rb") as fin:
            return signal_type.get_index_cls().deserialize(fin)


class CliSimpleState(helpers.SimpleFetchedStateStore):
    """
    A simple on-disk storage format for the CLI.

    Ideally, it should be easy to read manually (for debugging),
    but compact enough to handle very large sets of data.
    """

    def __init__(
        self, api_cls: t.Type[SignalExchangeAPI], fetched_state_dir: pathlib.Path
    ) -> None:
        super().__init__(api_cls)
        self.dir = fetched_state_dir

    def collab_file(self, collab_name: str) -> pathlib.Path:
        """The file location for collaboration state"""
        return self.dir / f"{collab_name}.state.pickle"

    def exists(self, collab: CollaborationConfigBase) -> bool:
        """
        Returns true if the state file is available

        This usually means that state is available, but the file could also be
        corrupt or unparsable.
        """
        return self.collab_file(collab.name).is_file()

    def clear(self, collab: CollaborationConfigBase) -> None:
        """Delete a collaboration and its state directory"""
        super().clear(collab)
        file = self.collab_file(collab.name)
        if file.is_file():
            logging.info("Removing %s", file)
            file.unlink(missing_ok=True)
        if file.parent.is_dir():
            if next(file.parent.iterdir(), None) is None:
                logging.info("Removing directory %s", file.parent)
                file.parent.rmdir()

    def _read_state(
        self,
        collab_name: str,
    ) -> t.Optional[FetchDeltaTyped]:
        file = self.collab_file(collab_name)
        if not file.is_file():
            return None
        try:
            with file.open("rb") as f:
                delta = pickle.load(f)

            assert isinstance(delta, FetchDelta), "Unexpected class type?"
            assert (
                delta.checkpoint.__class__ == self.api_cls.get_checkpoint_cls()
            ), f"wrong checkpoint class stored in {file}?"

            logging.debug("Loaded %s with %d records", collab_name, len(delta.updates))
            return delta
        except Exception:
            logging.exception("Failed to read state for %s", collab_name)
            raise CommandError(
                f"Failed to read state for {collab_name}. "
                "You might have to delete it with `threatexchange fetch --clear`"
            )

    def _write_state(
        self,
        collab_name: str,
        delta: FetchDeltaTyped,
    ) -> None:
        file = self.collab_file(collab_name)
        if not file.parent.exists():
            file.parent.mkdir(parents=True)

        if issubclass(self.api_cls, SignalExchangeAPIWithSimpleUpdates):
            record_sanity_check = next(
                (record for record in delta.updates.values()),
                None,
            )

            if record_sanity_check is not None:
                assert (  # Not isinstance - we want exactly this class
                    record_sanity_check.__class__ == self.api_cls.get_record_cls()
                ), (
                    f"Record cls: want {self.api_cls.get_record_cls().__name__} "
                    f"got {record_sanity_check.__class__.__name__}"
                )
        tmpfile = file.with_name(f".{file.name}")
        with tmpfile.open("wb") as f:
            pickle.dump(delta, f)
        tmpfile.replace(file)



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Abstraction around efficient scaled matching of SignalTypes

This usually involves build an efficient datastructure for lookups
(aka, an index).

At scale, the flow for matching looks something like:
1. Fetch information from API, store it locally
2. Process local copy of the information, separate it out by
   which system will consume it
3. Build indices for distribution (or deltas to those indices if
   you are really fancy)
4. In the service that does matching, swap out the index to the
   new one.
5. Hash content at upload sites, check against the service hosting
   the index.

"""

from dataclasses import dataclass
import pickle
import typing as t

T = t.TypeVar("T")
S_Co = t.TypeVar("S_Co", covariant=True, bound="SignalSimilarityInfo")
CT = t.TypeVar("CT", bound="Comparable")


class Comparable(t.Protocol):
    """Helper for annotating comparable types."""

    def __lt__(self: CT, other: CT) -> bool:
        pass

    def __le__(self: CT, other: CT) -> bool:
        pass


NO_DISTANCE_STR = "-"


class SignalSimilarityInfo:
    """
    Metadata with context about a comparison between content or signals.

    This can be used for logging and debugging, but could also be re-used
    as an argument to match functions to use as thresholds.

    This information often is treated as a distance, and so comparison
    operators
    """

    def __lt__(self, other: t.Any) -> bool:
        if isinstance(other, SignalSimilarityInfoWithSingleDistance):
            return False  # < operator valid, but by default, not ordered
        return NotImplemented

    def __le__(self, other: t.Any) -> bool:
        # Provide a default impl
        return self < other or self == other

    # Don't define __gt__ or __ge__ because of unexpected interactions with
    # functools.total_ordering
    def pretty_str(self) -> str:
        """
        A short string without whitespace about a match with more context.

        See it in action on the CLI in `threatexchange match`.
        """
        return NO_DISTANCE_STR


@dataclass
# @functools.total_ordering
# Can't use yet, need to move library mypy past
# https://github.com/python/mypy/issues/11728
class SignalSimilarityInfoWithSingleDistance(t.Generic[CT], SignalSimilarityInfo):
    distance: CT

    def pretty_str(self) -> str:
        return str(self.distance)

    def _comparable(
        self, other: t.Any
    ) -> t.Optional["SignalSimilarityInfoWithSingleDistance[CT]"]:
        if isinstance(other, SignalSimilarityInfoWithSingleDistance):
            if isinstance(self.distance, other.distance.__class__):
                return other
        return None

    def __lt__(self, other: t.Any) -> bool:
        checked = self._comparable(other)
        if checked is None:
            return NotImplemented
        return self.distance < checked.distance

    def __eq__(self, other: object) -> bool:
        checked = self._comparable(other)
        if checked is None:
            return super().__eq__(other)
        return self.distance == checked.distance


SignalSimilarityInfoWithIntDistance = SignalSimilarityInfoWithSingleDistance[int]


class IndexMatchUntyped(t.Generic[S_Co, T]):
    """
    Wrapper around a match to the index, which may or may not
    be an actual match based on the distance settings for a source.

    The index is free to optimize not returning values that are
    greater than would be considered matches by a downstream (i.e.
    31 is probably the max distance that would be used in PDQ)

    NamedTuple can't be made generic, so here's an equivalent
    """

    __slots__ = ["similarity_info", "metadata"]
    similarity_info: S_Co
    metadata: T

    def __init__(self, similarity_info: S_Co, metadata: T) -> None:
        self.similarity_info = similarity_info
        self.metadata = metadata

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, IndexMatchUntyped):
            return (
                self.similarity_info == other.similarity_info
                and self.metadata == other.metadata
            )
        return super().__eq__(other)


IndexMatch = IndexMatchUntyped[SignalSimilarityInfo, T]


Self = t.TypeVar("Self", bound="SignalTypeIndex")


class SignalTypeIndex(t.Generic[T]):
    """
    Abstraction for efficient scale matching on signals.

    SignalType forces you to implement the brute force approaches to
    make them easier to work with and conceptualize. However, if you
    are intending to match thousands of items a second at the scale of
    millions of signals, this class allows you to implement a more
    complex and efficient method.

    This class can be thought of as just a Dict[hash, List[T]] interface
    that can return the flattened result of multiple entries on a get().

    The T can be whatever metadata might be useful, although be warned
    that not all T may serialize correctly.

    # Handling Index Types that cannot be updated
    The default interface assumes that you can both create and update
    indices. It's possible that some storage types cannot support this.
    In that case build() and deserialize() are the only methods that
    should be used to create indices, and you can have add() methods
    throw NotImplementedError.

    # Handling Restricted Value Types
    In cases where your underlying index has limited type support for
    values (i.e. FAISS seems to only support ints), the easiest workaround
    is to just generate your own internal ids and maintain a separate map of
    the real results.

    fancy_index_only_to_int[str, int]
    internal_mapping List[T] or Dict[int, T]

    def get(hash: str):
        matches = fancy_index_only_to_int[str]
        ret = []
        for idx in matches:
            entry = internal_mapping[idx]
            ret.append(entry)
        return ret

    def add(hash: str, meta: T):
        idx = fancy_index_only_to_int.put(hash)
        internal_mapping[idx] = meta
    """

    # TODO - this doesn't handle bytes queries / BytesHashers
    def query(self, query: str) -> t.Sequence[IndexMatch[T]]:
        """
        Look up entries against the index, up to the max supported distance.

        The index is free to optimize not returning values that are
        greater than would be considered matches by a downstream system
        (i.e. 31 is probably the max distance that would be used in PDQ)
        """
        raise NotImplementedError

    @classmethod
    def build(cls: t.Type[Self], entries: t.Iterable[t.Tuple[str, T]]) -> Self:
        """
        Build an index from a set of entries.

        You can override __init__ as needed, but keep this once constant.

        Note that there may be duplicates of the hash type, i.e.

        H1: M1
        H1: M2

        A later call to query(H1) should return both M1 and M2
        """
        ret = cls()
        ret.add_all(entries)
        return ret

    # TODO - probably move add() methods to a mixin instead
    def add(self, signal_str: str, entry: T) -> None:
        """
        Add an entry to the index.

        Duplicate entries should not replace previous ones.
        """
        raise NotImplementedError

    def add_all(self, entries: t.Iterable[t.Tuple[str, T]]) -> None:
        """add, but more so"""
        for signal_str, entry in entries:
            self.add(signal_str, entry)

    def serialize(self, fout: t.BinaryIO) -> None:
        """
        Convert the index into a bytestream (probably a file).

        The reason to use an IO instead of just stringifying it is that some
        indices might be really big, and we don't want to pay a cost of 2x-ing
        our memory storage (versus writing to file).

        Could also be premature optimization, you decide!
        """
        pickle.dump(self, fout, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def deserialize(cls: t.Type[Self], fin: t.BinaryIO) -> Self:
        """Instantiate an index from a previous call to serialize"""
        return pickle.load(fin)

    def dispose(self) -> None:
        """
        Release any native resources or large auxiliary structures held by the
        index implementation. Default is a no-op; implementations that wrap
        native libraries (e.g., FAISS) should override.
        """
        return None



# Copyright (c) Meta Platforms, Inc. and affiliates.

from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.extensions.vpdq.vpdq import VPDQSignal

TX_MANIFEST = ThreatExchangeExtensionManifest(
    signal_types=(VPDQSignal,),
)



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""Classify command for content moderation."""

import argparse
import os
import sys
import typing as t

from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli.exceptions import CommandError
from threatexchange.cli import command_base
from threatexchange.classifier.openai_moderation import (
    OpenAIModerationClassifier,
    MissingAPIKeyError,
)


class ClassifyCommand(command_base.CommandWithSubcommands):
    """Classify content using moderation APIs."""

    _SUBCOMMANDS: t.List[t.Type[command_base.Command]] = []

    @classmethod
    def get_name(cls) -> str:
        return "classify"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        pass


class ClassifyTextCommand(command_base.Command):
    """Classify text content using moderation APIs. Requires OPENAI_API_KEY."""

    @classmethod
    def get_name(cls) -> str:
        return "text"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "input",
            help="text string, file path, or - for stdin",
        )
        ap.add_argument(
            "--mod-api",
            action="store_true",
            help="use OpenAI Moderation API (default)",
        )
        ap.add_argument(
            "-a", "--show-all", action="store_true", help="show all categories"
        )
        ap.add_argument(
            "-m",
            "--model",
            default="omni-moderation-latest",
            help="model to use (default: omni-moderation-latest)",
        )

    def __init__(
        self,
        input: str,
        mod_api: bool = False,
        show_all: bool = False,
        model: str = "omni-moderation-latest",
    ):
        self.input = input
        self.mod_api = mod_api
        self.show_all = show_all
        self.model = model

    def execute(self, settings: CLISettings) -> None:
        # Resolve text input
        if self.input == "-":
            text = sys.stdin.read()
        elif os.path.isfile(self.input):
            with open(self.input) as f:
                text = f.read()
        else:
            text = self.input

        # Default to mod-api if no API flag specified
        # (currently only mod-api is supported)
        try:
            classifier = OpenAIModerationClassifier(model=self.model)
        except MissingAPIKeyError as e:
            raise CommandError.user(str(e)) from e

        try:
            result = classifier.classify_text(text)
        except Exception as e:
            raise CommandError.external_dependency(f"API error: {e}") from e

        _print_classification_result(result, self.show_all)


class ClassifyImageCommand(command_base.Command):
    """Classify image content using moderation APIs."""

    @classmethod
    def get_name(cls) -> str:
        return "image"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "input",
            help="image file path or URL",
        )
        ap.add_argument(
            "--mod-api",
            action="store_true",
            help="use OpenAI Moderation API (default)",
        )
        ap.add_argument(
            "-a", "--show-all", action="store_true", help="show all categories"
        )
        ap.add_argument(
            "-m",
            "--model",
            default="omni-moderation-latest",
            help="model to use (default: omni-moderation-latest)",
        )

    def __init__(
        self,
        input: str,
        mod_api: bool = False,
        show_all: bool = False,
        model: str = "omni-moderation-latest",
    ):
        self.input = input
        self.mod_api = mod_api
        self.show_all = show_all
        self.model = model

    def execute(self, settings: CLISettings) -> None:
        raise CommandError.user("Image classification is not implemented yet")


class ClassifyHybridCommand(command_base.Command):
    """Classify image + text together (multi-modal) using moderation APIs."""

    @classmethod
    def get_name(cls) -> str:
        return "hybrid"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "image",
            help="image file path or URL",
        )
        ap.add_argument(
            "text",
            help="text caption or description for the image",
        )
        ap.add_argument(
            "--mod-api",
            action="store_true",
            help="use OpenAI Moderation API (default)",
        )
        ap.add_argument(
            "-a", "--show-all", action="store_true", help="show all categories"
        )
        ap.add_argument(
            "-m",
            "--model",
            default="omni-moderation-latest",
            help="model to use (default: omni-moderation-latest)",
        )

    def __init__(
        self,
        image: str,
        text: str,
        mod_api: bool = False,
        show_all: bool = False,
        model: str = "omni-moderation-latest",
    ):
        self.image = image
        self.text = text
        self.mod_api = mod_api
        self.show_all = show_all
        self.model = model

    def execute(self, settings: CLISettings) -> None:
        raise CommandError.user("Hybrid classification is not implemented yet")


def _print_classification_result(result, show_all: bool) -> None:
    """Print classification results in a formatted table."""
    if show_all:
        items = sorted(result.labels.items())
    else:
        items = sorted((l, i) for l, i in result.labels.items() if i.is_match)

    if not items:
        print("None")
    else:
        max_len = max(len(l) for l, _ in items)
        for label, info in items:
            flag = "FLAGGED" if info.is_match else ""
            print(f"  {label:<{max_len}}  {info.score:>6.2%}  {flag}")


ClassifyCommand._SUBCOMMANDS = [
    ClassifyTextCommand,
    ClassifyImageCommand,
    ClassifyHybridCommand,
]



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Settings used to inform a SignalExchangeAPI what to fetch
"""

from dataclasses import dataclass
import typing as t


@dataclass
class CollaborationConfigBase:
    """
    Settings used to inform a SignalExchangeAPI what to fetch.

    Extend with any additional fields that you need to inform your API how
    and what to fetch.

    Management of persisting these is left to the specific platform
    (i.e. CLI or HMA).
    """

    name: str
    api: str  # Often a const for implementations
    # Whether to fetch/sync. Some implementations (like the CLI) may gate matching
    # to avoid waiting for an index rebuild to stop processing matches
    enabled: bool
    # TODO - eventually re-merge these back into collab config, or allow them as
    #        mixins
    # # Only fetch and index these types
    # only_signal_types: t.Set[str]
    # # Don't fetch and index these types
    # not_signal_types: t.Set[str]
    # # Only use signals from these owners
    # only_owners: t.Set[int]
    # not_owners: t.Set[int]
    # # Only use signals with these tags
    # only_tags: t.Set[str]
    # not_tags: t.Set[str]


@dataclass
class CollaborationConfigWithDefaults(CollaborationConfigBase):
    """
    Helper to re-order arguments for defaults
    """

    api: str = ""
    enabled: bool = True
    # only_signal_types: t.Set[str] = field(default_factory=set)
    # not_signal_types: t.Set[str] = field(default_factory=set)
    # only_owners: t.Set[int] = field(default_factory=set)
    # not_owners: t.Set[int] = field(default_factory=set)
    # only_tags: t.Set[str] = field(default_factory=set)
    # not_tags: t.Set[str] = field(default_factory=set)


class CollaborationConfigStoreBase:
    def get_all_collabs(self) -> t.List[CollaborationConfigBase]:
        """
        Get all CollaborationConfigs, already resolved to the correct type
        """
        raise NotImplementedError

    def get_collab(self, name: str) -> t.Optional[CollaborationConfigBase]:
        """Get a specific collab config by name"""
        return next((c for c in self.get_all_collabs() if c.name == name), None)



# Copyright (c) Meta Platforms, Inc. and affiliates.

import vpdq
import faiss
from threatexchange.extensions.vpdq.vpdq_util import VpdqCompactFeature
from threatexchange.signal_type.pdq.pdq_utils import BITS_IN_PDQ
import typing as t
import numpy
import binascii
import weakref


class VPDQHashIndex:
    """Wrapper around an faiss binary index for use with searching for similar VPDQ features"""

    def __init__(self, faiss_index: t.Optional[faiss.Index] = None) -> None:
        """
        If none faiss index is provided, will use "brute-force" faiss search
        """
        self.faiss_index = (
            faiss.IndexBinaryFlat(BITS_IN_PDQ) if faiss_index is None else faiss_index
        )
        self._finalizer = weakref.finalize(
            self, VPDQHashIndex._finalize_faiss, self.faiss_index
        )

    def add_single_video(self, hashes: t.List[VpdqCompactFeature]) -> None:
        """
        Args:
            hashes : One video's VPDQ features of to create the index with
        """
        hash_bytes = [binascii.unhexlify(h.pdq_hex) for h in hashes]
        vectors = [numpy.frombuffer(h, dtype=numpy.uint8) for h in hash_bytes]
        self.faiss_index.add(numpy.array(vectors))

    def search_with_distance_in_result(
        self, queries: t.List[VpdqCompactFeature], distance_tolerance: int
    ) -> t.Dict[str, t.List[t.Tuple[int, int]]]:
        """
        Searches this index for PDQ hashes within the index that are no more than the threshold away
        from the query hashes by hamming distance.

        Args:
            queries : The VPDQ features to against the index
            distance_tolerance : Threshold value to use for this search. The hamming distance between the result hashes
            and the related query will be no more than the threshold value. i.e., hamming_dist(q_i,r_i_j) <= threshold.

        Returns:
        sequence of matches per query
            For each query provided in queries, the returned sequence will contain a sequence of matches within the index
            that were within threshold hamming distance of that query. These matches will be (idx, distance).
            The inner sequences may be empty in the case of no hashes within the index.
            The same VPDQ feature may also appear in more than one inner sequence if it matches multiple query hashes.
            For example the hash "000000000000000000000000000000000000000000000000000000000000ffff" would match both
            "00000000000000000000000000000000000000000000000000000000fffffff" and
            "0000000000000000000000000000000000000000000000000000000000000000" for a threshold of 16. Thus it would appear in
            the entry for both the hashes if they were both in the queries list.

            e.g.
            query_str =>  (idx, distance)
            result = {
                "000000000000000000000000000000000000000000000000000000000000ffff": [
                    (12345678901, 16.0)
                ]
            }
        """

        query_vectors = [
            numpy.frombuffer(binascii.unhexlify(q.pdq_hex), dtype=numpy.uint8)
            for q in queries
        ]
        qs = numpy.array(query_vectors)
        limits, similarities, neighbors = self.faiss_index.range_search(
            qs, distance_tolerance + 1
        )

        result = {}
        for i, query in enumerate(queries):
            matches = [idx.item() for idx in neighbors[limits[i] : limits[i + 1]]]
            distances = [idx for idx in similarities[limits[i] : limits[i + 1]]]
            result[query.pdq_hex] = list(zip(matches, distances))
        return result

    def __getstate__(self):
        data = faiss.serialize_index_binary(self.faiss_index)
        return data

    def __setstate__(self, data):
        self.faiss_index = faiss.deserialize_index_binary(data)
        # Re-register finalizer after unpickling
        self._finalizer = weakref.finalize(
            self, VPDQHashIndex._finalize_faiss, self.faiss_index
        )

    def dispose(self) -> None:
        try:
            reset_fn = getattr(self.faiss_index, "reset", None)
            if callable(reset_fn):
                reset_fn()
        except Exception:
            pass

    @staticmethod
    def _finalize_faiss(faiss_index: faiss.Index) -> None:
        try:
            reset_fn = getattr(faiss_index, "reset", None)
            if callable(reset_fn):
                reset_fn()
        except Exception:
            pass



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
SignalExchangeAPI impl for the NCMEC hash exchange API

@see NCMECSignalExchangeAPI
"""

import logging
import time
import typing as t
from dataclasses import dataclass, field

from threatexchange.exchanges.clients.ncmec import hash_api as api

from threatexchange.exchanges import auth, fetch_state as state
from threatexchange.exchanges import signal_exchange_api
from threatexchange.exchanges.collab_config import (
    CollaborationConfigWithDefaults,
)
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.signal_type.md5 import VideoMD5Signal
from threatexchange.signal_type.pdq.signal import PdqSignal

_API_NAME: str = "ncmec"


@dataclass
class _NCMECPagingInfo:
    """
    Store paging information for resume mid-page.

    NCMEC suggests not storing paging_urls long term so we consider them invalid
    12hr after the last_fetch_time
    """

    paging_url: str
    paging_end_ts: int
    paging_url_ts: int = field(default_factory=lambda: int(time.time()))

    @property
    def is_valid(self) -> bool:
        return time.time() - self.paging_url_ts <= 12 * 60 * 60

    def __bool__(self) -> bool:
        return self.is_valid


@dataclass
class NCMECCheckpoint(
    state.FetchCheckpointBase,
):
    """
    NCMEC primarily revolves around polling the timestamp.

    NCMEC IDs seem to stay around forever, so no need for is_stale()
    """

    # The biggest value of "to", and the next "from"
    get_entries_max_ts: int
    # A url to fetch the next page of results
    # Only reference this value through `paging_url` property
    paging_info: t.Optional[_NCMECPagingInfo] = None

    def get_progress_timestamp(self) -> t.Optional[int]:
        return self.get_entries_max_ts

    @classmethod
    def from_completed_ncmec_fetch(cls, target_end_ts: int) -> "NCMECCheckpoint":
        """Get the value for a completed fetch (iterated to no more records)"""
        return cls(target_end_ts)

    @classmethod
    def from_paged_ncmec_fetch(
        cls,
        response: api.GetEntriesResponse,
        *,
        current_start: int,
        current_end: int,
    ) -> "NCMECCheckpoint":
        """Get a checkpoint for an in-progress paged fetch"""
        if not response.next:  # This is actually a completed fetch
            return cls.from_completed_ncmec_fetch(current_end)
        return cls(
            current_start,
            _NCMECPagingInfo(paging_url=response.next, paging_end_ts=current_end),
        )

    def __setstate__(self, d: t.Dict[str, t.Any]) -> None:
        """Implemented for pickle version compatibility."""
        # 0.99.0 => 1.0.0:
        ### field 'max_timestamp' renamed to 'get_entries_max_ts'
        if "max_timestamp" in d:
            d["get_entries_max_ts"] = d.pop("max_timestamp")

        self.__dict__ = d


@dataclass
class _NCMECCollabConfigRequiredFields:
    environment: api.NCMECEnvironment = field(
        metadata={"help": "which database to connect to"}
    )


@dataclass
class NCMECCollabConfig(
    CollaborationConfigWithDefaults,
    _NCMECCollabConfigRequiredFields,
):
    api: str = field(init=False, default=_API_NAME)
    only_esp_ids: t.Set[int] = field(
        default_factory=set,
        metadata={
            "help": "Only take entries from these eletronic service provider (ESP) ids"
        },
    )


@dataclass
class NCMECOpinion(state.SignalOpinion):
    esp_id: int

    def __setstate__(self, d: t.Dict[str, t.Any]) -> None:
        """Implemented for pickle version compatibility."""
        # 0.99.0 => 1.0.0:
        ### field 'owner_id' renamed to 'esp_id'
        if "owner" in d:
            d["esp_id"] = d["owner"]
        super().__setstate__(d)


@dataclass
class NCMECSignalMetadata(state.FetchedSignalMetadata):
    """
    NCMEC metadata includes who uploaded it, as well as what they tagged.

    The NCMEC API has no concept of false positives - every entry is reported.
    """

    member_entries: t.Dict[int, t.Set[str]]

    def get_as_opinions(self) -> t.Sequence[NCMECOpinion]:
        return [
            NCMECOpinion(
                False,  # TODO - get my own esp_id
                state.SignalOpinionCategory.POSITIVE_CLASS,
                tags,
                member_id,
            )
            for member_id, tags in self.member_entries.items()
        ]


def _get_conversion(
    signal_types: t.Sequence[t.Type[SignalType]],
) -> t.Mapping[t.Tuple[api.NCMECEntryType, str], t.Type[SignalType]]:
    ret: t.Dict[t.Tuple[api.NCMECEntryType, str], t.Type[SignalType]] = {}
    if VideoMD5Signal in signal_types:
        ret[api.NCMECEntryType.video, "md5"] = VideoMD5Signal
    if PdqSignal in signal_types:
        ret[api.NCMECEntryType.image, "pdq"] = PdqSignal
    for st in signal_types:
        if st.get_name() == "photodna":
            ret[api.NCMECEntryType.image, "pdna"] = st
            break
    return ret


@dataclass
class NCMECCredentials(auth.CredentialHelper):
    ENV_VARIABLE: t.ClassVar[str] = "TX_NCMEC_CREDENTIALS"
    FILE_NAME: t.ClassVar[str] = "~/.tx_ncmec_credentials"

    user: str = field(metadata={"help": "NCMEC hash sharing API username."})
    password: str = field(metadata={"help": "NCMEC hash sharing API password."})

    @classmethod
    def _from_str(cls, s: str) -> "NCMECCredentials":
        user, _, passw = s.strip().partition(":")
        return cls(user, passw)

    def _are_valid(self) -> bool:
        return bool(self.user and self.password)


class NCMECSignalExchangeAPI(
    auth.SignalExchangeWithAuth[NCMECCollabConfig, NCMECCredentials],
    signal_exchange_api.SignalExchangeAPI[
        NCMECCollabConfig,
        NCMECCheckpoint,
        NCMECSignalMetadata,
        str,
        api.NCMECEntryUpdate,
    ],
):
    """
    Conversion for the NCMEC hash API

    Key implementation details:
        1. API is a stream of content: opinion, hashes,
           which need to be remapped to hash => opinion
        2. Owners have ids
        3. As of 5/2022 there are no false positive or seen statuses
    """

    MAX_FETCH_SIZE: t.ClassVar[int] = 400000
    FETCH_SHRINK_FACTOR: t.ClassVar[int] = 4

    def __init__(
        self,
        collab: NCMECCollabConfig,
        username: str,
        password: str,
    ) -> None:
        super().__init__()
        self.collab = collab
        self._username = username
        self._password = password

    @classmethod
    def for_collab(
        cls,
        collab: NCMECCollabConfig,
        credentials: t.Optional["NCMECCredentials"] = None,
    ) -> "NCMECSignalExchangeAPI":
        credentials = credentials or NCMECCredentials.get(cls)
        return cls(collab, credentials.user, credentials.password)

    @classmethod
    def get_name(cls) -> str:
        return _API_NAME

    @staticmethod
    def get_config_cls() -> t.Type[NCMECCollabConfig]:
        return NCMECCollabConfig

    @staticmethod
    def get_checkpoint_cls() -> t.Type[NCMECCheckpoint]:
        return NCMECCheckpoint

    @staticmethod
    def get_record_cls() -> t.Type[NCMECSignalMetadata]:
        return NCMECSignalMetadata

    @staticmethod
    def get_credential_cls() -> t.Type[NCMECCredentials]:
        return NCMECCredentials

    def get_client(self, environment: api.NCMECEnvironment) -> api.NCMECHashAPI:
        if not api.is_valid_user_pass(self._username, self._password):
            raise Exception("NCMEC username and password not configured or invalid.")
        return api.NCMECHashAPI(self._username, self._password, environment)

    def fetch_iter(
        self,
        _supported_signal_types: t.Sequence[t.Type[SignalType]],
        checkpoint: t.Optional[NCMECCheckpoint],
    ) -> t.Iterator[state.FetchDelta[str, api.NCMECEntryUpdate, NCMECCheckpoint]]:
        """
        Use flow control to efficiently fetch from the NCMEC API.

        The NCMEC API does not provide entries in a strictly ascending time order.
        As a result, the only checkpoint we can safely return is if we have
        exhausted an entire time range of fetching.

        However, the data within the NCMEC hash dbs are randomly distributed,
        with the following observed behavior:
            1. Fetching an empty range (0 entries) is fast
            2. We can estimate the number of entries in a range based on the
               parameters of the "next" field
            3. Fetching from a cursor of data is significantly faster than
               generating a new cursor.

        Therefore, we can aim for the following strategy:
            1. Prefer fetching empty ranges repeatedly rather than potentially
               generating an overfetch. (Shrink quickly, grow slowly)
            2. Be very generous with overfetch limits, since if we've generated
               the cursor
        """
        start_time = 0
        checkpoint_paging_info = None
        if checkpoint is not None:
            start_time = checkpoint.get_entries_max_ts
            checkpoint_paging_info = checkpoint.paging_info
        # Avoid being exactly at end time for updates showing up multiple
        # times in the fetch, since entries are not ordered by time
        end_time = int(time.time()) - 5

        client = self.get_client(self.collab.environment)
        # We could probably mutate start time, but new variable for clarity
        current_start = start_time
        # The range we are fetching
        duration = end_time - current_start
        # A counter for when we want to increase our duration
        # We want to be conservative
        low_fetch_counter = 0
        hi_fetch_counter = 0

        def duration_str(sec):
            if sec < 1800:
                return f"{sec} seconds"
            elif sec < 43200:
                return f"{sec / 3600:.2f} hours"
            return f"{sec / (3600 * 24):.2f} days"

        def log(event: str) -> None:
            """Helper to log fetch events"""
            logging.info(
                "NCMEC API %s @%s (%s)",
                event,
                api._date_format(current_start),
                duration_str(duration),
            )

        while current_start < end_time:  # We have not completed the interval
            # Duration updated by probing behavior
            duration = max(1, duration)  # Infinite loop defense
            # Don't fetch past our designated end
            current_end = min(end_time, current_start + duration)
            resume_paging_url = ""
            total_fetched = 0
            enumerate_start = 0
            # Use the checkpoint paging info exactly once
            if checkpoint and checkpoint_paging_info:
                assert checkpoint.get_entries_max_ts == current_start  # sanity
                assert checkpoint_paging_info.paging_end_ts < end_time
                resume_paging_url = checkpoint_paging_info.paging_url
                current_end = checkpoint_paging_info.paging_end_ts
                duration = current_end - current_start  # For logging string
                log("Resuming mid-page")
                enumerate_start = 1  # Skip over probing behavior
                checkpoint_paging_info = None
            for i, entry in enumerate(
                client.get_entries_iter(
                    start_timestamp=current_start,
                    end_timestamp=current_end,
                    resume_paging_url=resume_paging_url,
                ),
                start=enumerate_start,
            ):
                total_fetched += len(entry.updates)
                if (
                    i == 0 and duration > 1 and entry.next
                ):  # First fetch, do probing behavior
                    duration_guess = entry.max_timestamp - current_start
                    if (
                        duration > 60
                        or duration_guess < duration // self.FETCH_SHRINK_FACTOR
                    ):
                        old_duration = duration
                        # We want to at last shrink by our shrink factor
                        duration = min(
                            duration
                            // self.FETCH_SHRINK_FACTOR
                            // (2**hi_fetch_counter),
                            duration_guess - 1,
                        )
                        log(
                            f"est {duration_str(duration_guess)} span, "
                            + f"shrinking from {duration_str(old_duration)}"
                        )
                        hi_fetch_counter += 1
                        low_fetch_counter = 0  # Don't grow right after a shrink
                        break  # Retry get_entries_iter with new parameters
                else:
                    hi_fetch_counter = 0

                if i % 100 == 5:
                    # On large fetches, log notice every once in a while
                    log(f"large fetch ({i}) with {total_fetched} updates.")

                yield state.FetchDelta(
                    {f"{entry.member_id}-{entry.id}": entry for entry in entry.updates},
                    NCMECCheckpoint.from_paged_ncmec_fetch(
                        entry,
                        current_start=current_start,
                        current_end=current_end,
                    ),
                )

            else:  # Exhausted the fetch
                # If we're hovering near the single-fetch limit for a period
                # of time, we can likely safely expand our range.
                if total_fetched < api.NCMECHashAPI.ENTRIES_PER_FETCH * 2:
                    low_fetch_counter += 1
                    if low_fetch_counter >= self.FETCH_SHRINK_FACTOR:
                        log("multiple low fetches, increasing duration")
                        # Grow slower than we shrink
                        duration *= self.FETCH_SHRINK_FACTOR // 2
                        low_fetch_counter = 0
                # If we are not quite at our limit, but getting close to it,
                # pre-emptively shrink to try and stay under the limit
                elif total_fetched > self.MAX_FETCH_SIZE / self.FETCH_SHRINK_FACTOR:
                    log("close to overfetch limit, reducing duration")
                    duration //= self.FETCH_SHRINK_FACTOR
                    low_fetch_counter = 0
                else:  # Not too small, not too large, just right
                    low_fetch_counter = 0
                current_start = current_end

    @classmethod
    def fetch_value_merge(
        cls,
        old: t.Optional[api.NCMECEntryUpdate],
        new: t.Optional[api.NCMECEntryUpdate],
    ) -> t.Optional[api.NCMECEntryUpdate]:
        assert new is not None, "fetch shouldn't do this"
        if new.deleted:
            return None
        return new

    @classmethod
    def naive_convert_to_signal_type(
        cls,
        signal_types: t.Sequence[t.Type[SignalType]],
        collab: NCMECCollabConfig,
        fetched: t.Mapping[str, api.NCMECEntryUpdate],
    ) -> t.Dict[t.Type[SignalType], t.Dict[str, NCMECSignalMetadata]]:
        mapping: t.Mapping[t.Tuple[api.NCMECEntryType, str], t.Type[SignalType]] = (
            _get_conversion(signal_types)
        )
        ret: t.Dict[t.Type[SignalType], t.Dict[str, NCMECSignalMetadata]] = {}
        for entry in fetched.values():
            if entry.deleted:
                continue  # We expect len(fingerprints) == 0 here, but to be safe
            if collab.only_esp_ids and entry.member_id not in collab.only_esp_ids:
                continue
            for fingerprint_type, fingerprint_value in entry.fingerprints.items():
                st = mapping.get((entry.entry_type, fingerprint_type))
                if st is not None:
                    try:
                        signal_value = st.validate_signal_str(fingerprint_value)
                    except Exception:
                        logging.warning(
                            "Invalid fingerprint (%s): %s",
                            st.get_name(),
                            fingerprint_value,
                        )
                        continue
                    metadata = ret.setdefault(st, {}).setdefault(
                        signal_value, NCMECSignalMetadata({})
                    )
                    tags = metadata.member_entries.setdefault(entry.member_id, set())
                    if entry.classification:
                        tags.add(entry.classification)
        return ret



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Simple dbm implementation of the storage interface.
"""

import dbm
from enum import Enum
from dataclasses import dataclass
import typing as t
from pathlib import Path

from threatexchange.utils import dataclass_json
from threatexchange.storage import interfaces as iface
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.content_type.content_base import ContentType
from threatexchange.exchanges.signal_exchange_api import TSignalExchangeAPICls
from threatexchange.exchanges.impl.static_sample import StaticSampleSignalExchangeAPI
from threatexchange.signal_type.pdq.signal import PdqSignal
from threatexchange.signal_type.md5 import VideoMD5Signal
from threatexchange.content_type.photo import PhotoContent
from threatexchange.content_type.video import VideoContent


class _DbType(Enum):
    SIGNAL_TYPE = "signal_type"
    CONTENT_TYPE = "content_type"
    EXCHANGE_TYPE = "exchange_type"
    EXCHANGES = "exchanges"
    BANKS = "banks"
    INDEX = "index"


@dataclass
class _SignalTypeCfg:
    enabled_ratio: float = 1.0


# TODO - eventually to unified store
class DBMStore(iface.ISignalTypeConfigStore):
    """
    Local machine storage based on python dbm library,

    A simple one-machine storage interface, which persists on disk.
    Useful for testing or single machine work.
    """

    _folder: Path
    signal_types: t.Mapping[str, t.Type[SignalType]]
    content_types: t.Mapping[str, t.Type[ContentType]]
    exchange_types: t.Mapping[str, TSignalExchangeAPICls]

    def __init__(
        self,
        folder: Path,
        *,
        signal_types: t.Optional[t.Sequence[t.Type[SignalType]]] = None,
        content_types: t.Optional[t.Sequence[t.Type[ContentType]]] = None,
        exchange_types: t.Optional[t.Sequence[TSignalExchangeAPICls]] = None,
    ) -> None:
        assert folder.is_dir(), "must be path"
        self._folder = folder

        if signal_types is None:
            signal_types = [PdqSignal, VideoMD5Signal]
        if content_types is None:
            content_types = [PhotoContent, VideoContent]
        if exchange_types is None:
            exchange_types = t.cast(
                t.Sequence[TSignalExchangeAPICls], [StaticSampleSignalExchangeAPI]
            )

        self.signal_types = {st.get_name(): st for st in signal_types}
        self.content_types = {ct.get_name(): ct for ct in content_types}
        self.exchange_types = {et.get_name(): et for et in exchange_types}
        assert len(self.signal_types) == len(
            signal_types
        ), "All signal types must have unique names"
        assert len(self.content_types) == len(
            content_types
        ), "All content types must have unique names"
        assert len(self.exchange_types) == len(
            exchange_types
        ), "All exchange types must have unique names"

    def _open(self, db: _DbType, *, sub_db: str = ""):
        file = self._folder / str(db)
        if sub_db:
            file = file / sub_db
        return dbm.open(str(file), "c")

    def get_signal_type_configs(self) -> t.Mapping[str, iface.SignalTypeConfig]:
        """Return all installed signal types."""
        with self._open(_DbType.SIGNAL_TYPE) as db:
            ret = {}
            for name, st in self.signal_types.items():
                raw_cfg = db.get(name)
                cfg = _SignalTypeCfg()
                if raw_cfg is not None:
                    cfg = dataclass_json.dataclass_loads(
                        raw_cfg.decode(), _SignalTypeCfg
                    )
                ret[name] = iface.SignalTypeConfig(
                    signal_type=st,
                    enabled_ratio=cfg.enabled_ratio,
                )
        return ret

    def _create_or_update_signal_type_override(
        self, signal_type: str, enabled_ratio: float
    ) -> None:
        with self._open(_DbType.SIGNAL_TYPE) as db:
            db[signal_type] = dataclass_json.dataclass_dumps(
                _SignalTypeCfg(enabled_ratio=enabled_ratio)
            ).encode()



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""Simple implementation for the StopNCII REST API"""

from dataclasses import dataclass, field
import enum
import logging
import time
import typing as t

import dacite
import requests
from urllib3.util.retry import Retry
from threatexchange.exchanges.clients.utils.common import TimeoutHTTPAdapter


@enum.unique
class StopNCIISignalType(enum.Enum):
    """What the serialized hash represents"""

    Unknown = "Unknown"
    ImagePDQ = "ImagePDQ"
    VideoMD5 = "VideoMD5"
    VideoTMK = "VideoTMK"
    Text = "Text"
    URL = "URL"


@enum.unique
class StopNCIICaseStatus(enum.Enum):
    """The state of the user-submitted hash"""

    Unknown = "Unknown"
    Received = "Received"  # Initial state
    Active = "Active"  # Unclear what this represents
    Withdrawn = "Withdrawn"  # Should be deleted by client
    Deleted = "Deleted"  # Should be deleted by client


@enum.unique
class StopNCIICSPFeedbackValue(enum.Enum):
    """The feedback that a CSP has given on a hash"""

    Unknown = "Unknown"
    None_ = "None"  # Allows you to tag without associating a state
    QualityUnknown = "QualityUnknown"  # Unsure what this is used for
    PendingReview = "PendingReview"  # There was a match
    Blocked = "Blocked"  # Match reviewed, match was determined to be NCII
    NotBlocked = "NotBlocked"  # Match reviewed, inconclusive if NCII
    Withdrawn = "Withdrawn"  # Should be deleted by client
    Deleted = "Deleted"  # Should be deleted by client


@dataclass
class StopNCIICSPFeedback:
    """Feedback on a hash from a Content Service Provider"""

    feedbackValue: StopNCIICSPFeedbackValue  # What the feedback is
    tags: t.Set[str] = field(default_factory=set)  # Unstructured additional tags
    source: str = ""  # Name of the Content Service Provider (CSP)

    def as_dict_for_post(self) -> t.Dict[str, t.Any]:
        """The json-friendly format to send in post requests"""
        return {
            "tags": list(self.tags),
            "feedbackValue": self.feedbackValue.value,
        }


@dataclass
class StopNCIIHashRecord:
    """An aggregate record for a single hash from the StopNCII API"""

    lastModtimestamp: int  # Last update time
    signalType: StopNCIISignalType  # What the hashValue is
    hashValue: str  # The value of the hash
    hashStatus: StopNCIICaseStatus  # The "aggregate" case status
    caseNumbers: t.Dict[str, StopNCIICaseStatus]  # Individual cases that correspond
    # Feedback on hashes by CSPs
    CSPFeedbacks: t.List[StopNCIICSPFeedback] = field(default_factory=list)


@dataclass
class FetchHashesResponse:
    """
    Wrapper around the FetchHashes call response.

    Advantages of using a dataclass is the typing!
    """

    count: int  # How many records are there?
    nextPageToken: t.Optional[
        str
    ]  # Cursor for paginating, not valid over long periods of time
    nextSetTimestamp: int  # The best timestamp to use to store as a checkpoint
    hasMoreRecords: bool  # If the cursor is fully played out
    hashRecords: t.List[StopNCIIHashRecord]  # The records


@enum.unique
class StopNCIIEndpoint(enum.Enum):
    """Endpoints (with their own function keys) on StopNCII"""

    FetchHashes = "FetchHashes"
    SubmitHashes = "SubmitHashes"
    SubmitFeedback = "SubmitFeedback"


class StopNCIIAPI:
    """
    A wrapper around the StopNCII.org hash exchange API.

    Hashes are submitted by individual people to the portal at StopNCII.org,
    and Content Service Providers (CSPs), such as social networks, can provide
    feedback on the hashes on whether they are able to verify that the content
    corresponds to NCII content.
    """

    BASE_URL: t.ClassVar[str] = "https://api.stopncii.org/v1"

    DEFAULT_START_TIME: t.ClassVar[int] = 10

    def __init__(
        self,
        subscription_key: str,
        fetch_function_key: str,
        additional_function_keys: t.Optional[t.Dict[StopNCIIEndpoint, str]] = None,
        base_url_override: t.Optional[str] = None,
    ) -> None:
        self._function_keys = dict(additional_function_keys or {})
        self._function_keys[StopNCIIEndpoint.FetchHashes] = fetch_function_key
        self._subscription_key = subscription_key

        self._base_url = base_url_override or self.BASE_URL

    def _get_session(self, endpoint: StopNCIIEndpoint):
        """
        Custom requests sesson

        Ideally, should be used within a context manager:
        ```
        with self._get_session() as session:
            session.get()...
        ```

        If using without a context manager, ensure you end up calling close() on
        the returned value.
        """
        function_key = self._function_keys.get(endpoint)
        if not function_key:
            raise ValueError(
                f"You don't have a function key for the endpoint {endpoint}"
            )

        session = requests.Session()
        session.headers.update(
            {
                "x-functions-key": function_key,
                "Ocp-Apim-Subscription-Key": self._subscription_key,
            }
        )
        session.mount(
            self._base_url,
            adapter=TimeoutHTTPAdapter(
                timeout=60,
                max_retries=Retry(
                    total=4,
                    status_forcelist=[429, 500, 502, 503, 504],
                    # No retry for post. Could probably add timeout...
                    allowed_methods=["HEAD", "GET", "OPTIONS"],
                    backoff_factor=0.2,  # ~1.5 seconds of retries
                ),
            ),
        )
        return session

    def _get(self, endpoint: StopNCIIEndpoint, **params) -> t.Any:
        """
        Perform an HTTP GET request, and return the JSON response payload.

        Same timeouts and retry strategy as `_get_session` above.
        """

        url = "/".join((self._base_url, endpoint.value))
        with self._get_session(endpoint) as session:
            response = session.get(url, params=params)
            response.raise_for_status()
            return response.json()

    def _post(self, endpoint: StopNCIIEndpoint, *, json=None) -> t.Any:
        """
        Perform an HTTP POST request, and return the JSON response payload.

        No timeout or retry strategy.
        """

        url = "/".join((self._base_url, endpoint.value))
        with self._get_session(endpoint) as session:
            response = session.post(url, json=json)
            response.raise_for_status()
            return response.json()

    def fetch_hashes(
        self,
        *,
        page_size: int = 800,
        start_timestamp: int = DEFAULT_START_TIME,
        next_page: str = "",
    ) -> FetchHashesResponse:
        """
        Fetch a series of update records from the hash API.

        Records represent the current snapshot of all data, so if you see
        the same SignalType+Hash in a later iteration, it should completely
        replace the previously observed record.
        """
        params: t.Dict[str, t.Any] = {
            "startTimestamp": start_timestamp,
            "pageSize": page_size,
        }
        if next_page:
            params["nextPageToken"] = next_page
        logging.debug("StopNCII FetchHashes called: %s", params)
        json_val = self._get(StopNCIIEndpoint.FetchHashes, **params)
        logging.debug("StopNCII FetchHashes returns: %s", json_val)
        # If there is a malformed record or a change that would prevent the deserialization
        # of a record, skip over it instead of crashing. Please open an issue if you see the logging
        # statement below.
        records: t.List[StopNCIIHashRecord] = []
        for record in json_val.get("hashRecords", []):
            try:
                records.append(
                    dacite.from_dict(
                        data_class=StopNCIIHashRecord,
                        data=record,
                        config=dacite.Config(cast=[enum.Enum, set]),
                    )
                )
            except ValueError as e:
                logging.error(
                    "Convert response from JSON to Dataclass failed, err: %s, record: %s",
                    e,
                    record,
                )
        response: FetchHashesResponse = dacite.from_dict(
            data_class=FetchHashesResponse,
            data={**json_val, "hashRecords": []},
            config=dacite.Config(cast=[enum.Enum, set]),
        )
        response.hashRecords = records
        return response

    def fetch_hashes_iter(
        self, start_timestamp: int = DEFAULT_START_TIME
    ) -> t.Iterator[FetchHashesResponse]:
        """
        A simple wrapper around FetchHashes to keep fetching until complete.

        You could get the entire record stream from:
        everything = {
            (record.signalType, record.hashValue): record
            for result in api.fetch_hashes_iter()
            for record in result.hashRecords
        }

        """
        has_more = True
        next_page = ""
        while has_more:
            result = self.fetch_hashes(
                start_timestamp=start_timestamp, next_page=next_page
            )
            has_more = result.hasMoreRecords
            next_page = result.nextPageToken or ""
            yield result

    def submit_hash(
        self,
        signal_type: StopNCIISignalType,
        hash_value: str,
        tags: t.Optional[t.Set[str]] = None,
        reported_state: StopNCIICSPFeedbackValue = StopNCIICSPFeedbackValue.Blocked,
    ) -> None:
        """Convenience accessor for reporting a single hash"""
        self.submit_hashes(
            {
                (signal_type, hash_value): StopNCIICSPFeedback(
                    reported_state, tags or set()
                )
            }
        )

    def submit_hashes(
        self, hashes: t.Dict[t.Tuple[StopNCIISignalType, str], StopNCIICSPFeedback]
    ) -> None:
        """
        Upload hashes as a CSP to StopNCII.org.

        Most hashes come from users submitting to the portal, but the API
        allows for CSPs to source their own hashes as well.
        """
        now = int(time.time())
        records = []
        for (signal_type, hashValue), feedback in hashes.items():
            records.append(
                {
                    "lastModtimestamp": now,
                    "signalType": str(signal_type),
                    "hashValue": hashValue,
                    "CSPFeedbacks": [feedback.as_dict_for_post()],
                }
            )
        payload = {"count": len(hashes), "hashRecords": records}
        self._post(StopNCIIEndpoint.SubmitHashes, json=payload)

    def submit_feedback(
        self,
        signal_type: StopNCIISignalType,
        hash_value: str,
        feedback: StopNCIICSPFeedbackValue,
        tags: t.Optional[t.Set[str]] = None,
    ) -> None:
        """
        Convenience function for submitting feedback for a single hash.

        Note that even a single hash may correspond to multiple cases.
        """
        self.submit_feedbacks(
            {(signal_type, hash_value): StopNCIICSPFeedback(feedback, tags or set())}
        )

    def submit_feedbacks(
        self, feedbacks: t.Dict[t.Tuple[StopNCIISignalType, str], StopNCIICSPFeedback]
    ) -> None:
        """Submit feedback on multiple hashes"""
        now = int(time.time())
        # Now for the fun hacky part, convert to expected format
        feedback_dicts = []
        for (_signal_type, hashValue), feedback in feedbacks.items():
            # Why isn't signalType used here?
            fb_dict = feedback.as_dict_for_post()
            fb_dict["hashValue"] = hashValue
            feedback_dicts.append(fb_dict)
        payload = {"count": len(feedbacks), "hashFeedbacks": feedback_dicts}
        self._post(StopNCIIEndpoint.SubmitFeedback, json=payload)


def is_valid_key(key: str) -> bool:
    """
    Returns true if the string looks like a valid stopncii key
    """
    return bool(key)  # TODO



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
This is an entire copy of a file from ThreatExchange/hashing

TODO: Slim down to only what we need
"""

import copy
import json
import typing as t
import re

import urllib.parse
import urllib.error

import requests
from urllib3.util.retry import Retry


from .api_representations import ThreatPrivacyGroup
from threatexchange.exchanges.clients.utils.common import TimeoutHTTPAdapter


def is_valid_app_token(token: str) -> bool:
    """Returns true if the string looks like a valid token"""
    return bool(re.match("[0-9]{8,}(?:%7C|\\|)[a-zA-Z0-9_\\-]{20,}", token))


class _CursoredResponse:
    """Wrapper around paginated responses from Graph API"""

    def __init__(self, api: "ThreatExchangeAPI", url, params, decode_fn=None) -> None:
        self.api = api
        self.response = None
        self.next_url = url
        self.params = params
        self.data: t.List = []
        self.decode_fn = decode_fn

    @property
    def done(self):
        return self.next_url is None

    def next(self):
        if self.done:
            return []
        response = self.api.get_json_from_url(self.next_url, self.params)
        next_url = response.get("paging", {}).get("next")
        data = response.get("data", [])
        if self.decode_fn:
            data = [self.decode_fn(x) for x in data]
        self.next_url = next_url
        self.data = data
        self.params.clear()
        return self.data

    def __iter__(self):
        while not self.done:
            self.next()
            if self.data is not None:
                yield self.data


class ThreatExchangeAPI:
    _TE_BASE_URL = "https://graph.facebook.com/v21.0"

    # This is just a keystroke-saver / error-avoider for passing around
    # post-parameter field names.

    _POST_PARAM_NAMES = {
        "indicator": "indicator",  # For submit
        "type": "type",  # For submit
        "descriptor_id": "descriptor_id",  # For update
        "description": "description",
        "share_level": "share_level",
        "status": "status",
        "privacy_type": "privacy_type",
        "privacy_members": "privacy_members",
        "tags": "tags",
        "add_tags": "add_tags",
        "remove_tags": "remove_tags",
        "confidence": "confidence",
        "precision": "precision",
        "review_status": "review_status",
        "severity": "severity",
        "expired_on": "expired_on",
        "first_active": "first_active",
        "last_active": "last_active",
        "related_ids_for_upload": "related_ids_for_upload",
        "related_triples_for_upload_as_json": "related_triples_for_upload_as_json",
        # Legacy : should have been named reactions_to_add, but isn't. :(
        "reactions": "reactions",
        "reactions_to_remove": "reactions_to_remove",
    }

    def __init__(
        self,
        api_token: str,
        *,
        endpoint_override: t.Optional[str] = None,
    ) -> None:
        self.api_token = api_token
        self._base_url = endpoint_override or self._TE_BASE_URL

    @property
    def app_id(self):
        return int(self.api_token.partition("|")[0])

    def get_json_from_url(
        self, url, params=None, *, json_obj_hook: t.Optional[t.Callable] = None
    ):
        """
        Perform an HTTP GET request, and return the JSON response payload.
        Same timeouts and retry strategy as `_get_session` above.
        """
        with self._get_session() as _session:
            response = requests.get(url, params=params or {})  # !!! Typo? session.get?
            response.raise_for_status()
            return response.json(object_hook=json_obj_hook)

    def _get_session(self):
        """
        Custom requests sesson

        Ideally, should be used within a context manager:
        ```
        with self._get_session() as session:
            session.get()...
        ```

        If using without a context manager, ensure you end up calling close() on
        the returned value.
        """
        session = requests.Session()
        session.mount(
            self._base_url,
            adapter=TimeoutHTTPAdapter(
                timeout=60,
                max_retries=Retry(
                    total=4,
                    status_forcelist=[429, 500, 502, 503, 504],
                    allowed_methods=["HEAD", "GET", "OPTIONS"],
                    backoff_factor=0.2,  # ~1.5 seconds of retries
                ),
            ),
        )
        return session

    def get_tag_id(self, tagName, showURLs=False):
        """
        Looks up the "objective tag" ID for a given tag.
        This is suitable input for the /threat_tags endpoint.
        """
        url = (
            self._base_url
            + "/threat_tags"
            + "/?access_token="
            + self.api_token
            + "&text="
            + urllib.parse.quote(tagName)
        )
        if showURLs:
            print("URL:")
            print(url)

        response = self.get_json_from_url(url)

        # The lookup will get everything that has this as a prefix.
        # So we need to filter the results. This loop also handles the
        # case when the results array is empty.
        #
        # Example: when querying for "media_type_video", we want the 2nd one:
        # { "data": [
        #   { "id": "9999338563303771", "text": "media_type_video_long_hash" },
        #   { "id": "9999474908560728", "text": "media_type_video" },
        #   { "id": "9889872714202918", "text": "media_type_video_hash_long" }
        #   ], ...
        # }
        data = response["data"]
        desired = list(filter(lambda o: o["text"] == tagName, data))
        if len(desired) < 1:
            return None
        else:
            return desired[0]["id"]

    def get_threat_descriptors(self, ids, **kwargs):
        """
        Looks up all metadata for given IDs.
        """
        verbose = kwargs.get("verbose", False)
        showURLs = kwargs.get("showURLs", False)
        includeIndicatorInOutput = kwargs.get("includeIndicatorInOutput", True)

        default_fields = [
            "raw_indicator",
            "type",
            "added_on",
            "last_updated",
            "confidence",
            "owner",
            "privacy_type",
            "review_status",
            "status",
            "severity",
            "share_level",
            "tags",
            "description",
            "reactions",
            "my_reactions",
        ]
        fields = kwargs.get("fields", default_fields)

        # Check well-formattedness of descriptor IDs (which may have come from
        # arbitrary data on stdin).
        for id in ids:
            try:
                _ = int(id)
            except ValueError:
                raise Exception('Malformed descriptor ID "%s"' % id)

        # See also
        # https://developers.facebook.com/docs/threat-exchange/reference/apis/threat-descriptor/
        # for available fields

        url = (
            self._base_url
            + "/?access_token="
            + self.api_token
            + "&ids="
            + ",".join(ids)
            + "&fields="
            + ",".join(fields)
        )

        if showURLs:
            print("URL:")
            print(url)

        response = self.get_json_from_url(url)

        descriptors = []
        for id, descriptor in response.items():
            if not includeIndicatorInOutput:
                del descriptor["raw_indicator"]
            if verbose:
                print(json.dumps(descriptor))

            if "tags" in fields:
                # tags is returned as a dict sturctred like:
                # "tags": {
                #   "data": [
                #         {
                #            "id": "03465026013486502",
                #            "text": "uploaded_by_hma"
                #         }
                #      ]
                #   },
                #
                # Canonicalize the tag ordering and simplify the
                # structure to simply an array of tag-texts
                tags = descriptor.get("tags", {"data": []})["data"]
                descriptor["tags"] = sorted(tag["text"] for tag in tags)

            if descriptor.get("description") is None and "description" in fields:
                descriptor["description"] = ""

            descriptors.append(descriptor)
        return descriptors

    def get_threat_updates(
        self,
        privacy_group: int,
        *,
        start_time: t.Optional[int] = None,
        stop_time: t.Optional[int] = None,
        types: t.Iterable[str] = (),
        page_size: t.Optional[int] = None,
        fields: t.Optional[t.Iterable[str]] = None,
        decode_fn: t.Optional[t.Callable[[t.Any], t.Any]] = None,
    ) -> _CursoredResponse:
        """Gets threat updates for the given privacy group."""

        if fields is None:
            fields = (
                "id",
                "indicator",
                "type",
                "creation_time",
                "last_updated",
                "should_delete",
                "tags",
                "status",
                "applications_with_opinions",
            )

        params = {
            "access_token": self.api_token,
            "start_time": start_time,
            "stop_time": stop_time,
            "limit": page_size,
            "fields": ",".join(fields),
        }
        if types:
            params["types"] = ",".join(types)

        url = f"{self._base_url}/{privacy_group}/threat_updates/"
        return _CursoredResponse(self, url, params, decode_fn=decode_fn)

    def get_privacy_group(self, id: int) -> ThreatPrivacyGroup:
        """
        Returns a non-paginated list of all privacy groups the current app is a
        member of.
        """
        fields = [
            "id",
            "members_can_see",
            "members_can_use",
            "name",
            "description",
            "last_updated",
            "added_on",
            "threat_updates_enabled",
        ]
        url = self._get_graph_api_url(f"{id}", {"fields": ",".join(fields)})
        response = self.get_json_from_url(url)
        return ThreatPrivacyGroup.from_graph_api_dict(response)

    def get_threat_privacy_groups_member(
        self,
    ) -> t.List[ThreatPrivacyGroup]:
        """
        Returns a non-paginated list of all privacy groups the current app is a
        member of.
        """
        fields = [
            "id",
            "members_can_see",
            "members_can_use",
            "name",
            "description",
            "last_updated",
            "added_on",
            "threat_updates_enabled",
        ]
        url = self._get_graph_api_url(
            f"{self.app_id}/threat_privacy_groups_member", {"fields": ",".join(fields)}
        )
        response = self.get_json_from_url(url)
        return [ThreatPrivacyGroup.from_graph_api_dict(d) for d in response["data"]]

    def get_threat_privacy_groups_owner(
        self,
    ) -> t.List[ThreatPrivacyGroup]:
        """
        Returns a non-paginated list of all privacy groups the current app is a
        owner of.
        """
        fields = [
            "id",
            "members_can_see",
            "members_can_use",
            "name",
            "description",
            "last_updated",
            "added_on",
            "threat_updates_enabled",
        ]
        url = self._get_graph_api_url(
            f"{self.app_id}/threat_privacy_groups_owner", {"fields": ",".join(fields)}
        )
        response = self.get_json_from_url(url)
        return [ThreatPrivacyGroup.from_graph_api_dict(d) for d in response["data"]]

    def _get_graph_api_url(
        self, sub_path: t.Optional[str], query_dict: t.Dict = {}
    ) -> str:
        """
        Returns a threatexchange URL for a sub-path and a dictionary of query
        parameters. Automatically adds access_token to the query dictionary.
        """
        if "access_token" not in query_dict:
            query_dict["access_token"] = self.api_token

        query = urllib.parse.urlencode(query_dict)

        base_url_parts = urllib.parse.urlparse(self._base_url)
        url_parts = urllib.parse.ParseResult(
            base_url_parts.scheme,
            base_url_parts.netloc,
            f"{base_url_parts.path}/{sub_path}",
            base_url_parts.params,
            query,
            base_url_parts.fragment,
        )

        return urllib.parse.urlunparse(url_parts)

    def _validate_post_params_for_submit(self, postParams):
        """
        Returns error message or None.
        This simply checks to see (client-side) if required fields aren't provided.
        """
        if postParams.get(self._POST_PARAM_NAMES["descriptor_id"]) != None:
            return "descriptor_id must not be specified for submit."

        requiredFields = [
            self._POST_PARAM_NAMES["indicator"],
            self._POST_PARAM_NAMES["type"],
            self._POST_PARAM_NAMES["description"],
            self._POST_PARAM_NAMES["share_level"],
            self._POST_PARAM_NAMES["status"],
            self._POST_PARAM_NAMES["privacy_type"],
        ]

        missingFields = [
            fieldName if postParams.get(fieldName) == None else None
            for fieldName in requiredFields
        ]
        missingFields = [fieldName for fieldName in missingFields if fieldName != None]

        if len(missingFields) == 0:
            return None
        elif len(missingFields) == 1:
            return "Missing field %s" % missingFields[0]
        else:
            return "Missing fields %s" % ",".join(missingFields)

    def _validate_post_pararms_for_copy(self, postParams):
        """
        Returns error message or None.
        This simply checks to see (client-side) if required fields aren't provided.
        """
        if postParams.get(self._POST_PARAM_NAMES["descriptor_id"]) == None:
            return "Source-descriptor ID must be specified for copy."
        if postParams.get(self._POST_PARAM_NAMES["privacy_type"]) == None:
            return "Privacy type must be specified for copy."
        if postParams.get(self._POST_PARAM_NAMES["privacy_members"]) == None:
            return "Privacy members must be specified for copy."
        return None

    def react_to_threat_descriptor(
        self, descriptor_id, reaction, *, showURLs=False, dryRun=False
    ):
        """
        Does a POST to the reactions API.

        See: https://developers.facebook.com/docs/threat-exchange/reference/reacting
        """
        return self._postThreatDescriptor(
            "/".join(
                (
                    self._base_url,
                    str(descriptor_id),
                    f"?access_token={self.api_token}",
                )
            ),
            {"reactions": reaction},
            showURLs=showURLs,
            dryRun=dryRun,
        )

    def remove_reaction_from_threat_descriptor(
        self, descriptor_id, reaction, *, showURLs=False, dryRun=False
    ) -> t.List:
        """
        Does a POST to the reactions API.

        See: https://developers.facebook.com/docs/threat-exchange/reference/reacting
        """
        return self._postThreatDescriptor(
            "/".join(
                (
                    self._base_url,
                    str(descriptor_id),
                    f"?access_token={self.api_token}",
                )
            ),
            {"reactions_to_remove": reaction},
            showURLs=showURLs,
            dryRun=dryRun,
        )

    def react_matched_threat_descriptor(
        self, descriptor_id, *, showURLs=False, dryRun=False
    ):
        """
        Does a POST to the reactions API indicating that this descriptor's hash was
        matched.

        See: https://developers.facebook.com/docs/threat-exchange/reference/reacting
        """
        return self.react_to_threat_descriptor(
            descriptor_id, "SAW_THIS_TOO", showURLs=showURLs, dryRun=dryRun
        )

    def react_upvote_threat_descriptor(
        self, descriptor_id, *, showURLs=False, dryRun=False
    ):
        """
        Does a POST to the reactions API indicating that this descriptor's hash is
        helpful for discovering harmful content

        See: https://developers.facebook.com/docs/threat-exchange/reference/reacting
        """
        return self.react_to_threat_descriptor(
            descriptor_id, "HELPFUL", showURLs=showURLs, dryRun=dryRun
        )

    def react_downvote_threat_descriptor(
        self, descriptor_id, *, showURLs=False, dryRun=False
    ):
        """
        Does a POST to the reactions API indicating that this descriptor's hash is
        NOT helpful for discovering harmful content

        See: https://developers.facebook.com/docs/threat-exchange/reference/reacting
        """
        return self.react_to_threat_descriptor(
            descriptor_id, "NOT_HELPFUL", showURLs=showURLs, dryRun=dryRun
        )

    def upload_threat_descriptor(self, postParams, showURLs, dryRun):
        """
        Does a single POST to the threat_descriptors endpoint.  See also
        https://developers.facebook.com/docs/threat-exchange/reference/submitting
        """
        errorMessage = self._validate_post_params_for_submit(postParams)
        if errorMessage != None:
            return [errorMessage, None, None]

        url = "/".join(
            (self._base_url, "threat_descriptors", f"?access_token={self.api_token}")
        )

        return self._postThreatDescriptor(url, postParams, showURLs, dryRun)

    def copy_threat_descriptor(self, postParams, showURLs, dryRun):
        errorMessage = self._validate_post_pararms_for_copy(postParams)
        if errorMessage != None:
            return [errorMessage, None, None]

        # Get source descriptor
        sourceID = postParams["descriptor_id"]
        # Not valid for posting a new descriptor
        del postParams["descriptor_id"]
        sourceDescriptor = self.get_threat_descriptors([sourceID], showURLs=showURLs)
        sourceDescriptor = sourceDescriptor[0]

        # Mutate necessary fields
        newDescriptor = copy.deepcopy(sourceDescriptor)
        newDescriptor["indicator"] = sourceDescriptor["raw_indicator"]
        del newDescriptor["raw_indicator"]
        if "tags" in newDescriptor and newDescriptor["tags"] is None:
            del newDescriptor["tags"]

        # The shape is different between the copy-from data (mapping app IDs to
        # reactions) and the post data (just a comma-delimited string of owner-app
        # reactions).
        if "reactions" in newDescriptor:
            del newDescriptor["reactions"]

        # Take the source-descriptor values and overwrite any post-params fields
        # supplied by the caller. Note: Python's dict-update method keeps the old
        # value for a given field name when both old and new are present so we
        # invoke it seemingly 'backward'.
        #
        # Example:
        # * x = {'a': 1, 'b': 2, 'c': 3}
        # * y = {'a': 1, 'b': 9, 'd': 12}
        # * After y.update(x) then x is unchanged and y is
        #       {'a': 1, 'b': 2, 'd': 12, 'c': 3}
        #
        # This means we want newDescriptor.update(postParams)
        newDescriptor.update(postParams)

        # Get rid of fields like last_upated from the source descriptor which
        # aren't valid for post
        postParams = {}
        for key, value in newDescriptor.items():
            if self._POST_PARAM_NAMES.get(key) != None:
                postParams[key] = value

        return self.upload_threat_descriptor(postParams, showURLs, dryRun)

    def delete_threat_descriptor(
        self, descriptor_id, showURLs, dryRun
    ) -> t.List[t.Any]:
        url = (
            self._base_url
            + "/"
            + str(descriptor_id)
            + "?access_token="
            + self.api_token
        )
        if showURLs:
            print()
            print("(DELETE) URL:")
            print(url)
        if dryRun:
            print("Not doing DELETE since --dry-run.")
            return [None, None, ""]

        try:
            with self._get_session() as session:
                return [None, None, session.delete(url).json()]

        except urllib.error.HTTPError as e:
            responseBody = json.loads(e.read().decode("utf-8"))
            return [None, e, responseBody]

    def _postThreatDescriptor(self, url, postParams, showURLs, dryRun):
        """Code-reuse for submit and update"""
        for key, value in postParams.items():
            url += "&%s=%s" % (key, urllib.parse.quote(str(value)))
        if showURLs:
            print()
            print("(POST) URL:")
            print(url)
        if dryRun:
            print("Not doing POST since --dry-run.")
            return [None, None, ""]

        # Encode the inputs to the POST
        header = {"Content-Type": "text/json", "charset": "utf-8"}
        # This is a string
        data = urllib.parse.urlencode(postParams)
        # Turn it into a Python bytes object
        data = data.encode("ascii")

        # Do the POST
        try:
            with self._get_session() as session:
                return [None, None, session.post(url, data).json()]

        except urllib.error.HTTPError as e:
            responseBody = json.loads(e.read().decode("utf-8"))
            return [None, e, responseBody]

    def get_threat_descriptors_from_indicator(
        self, indicator_id: int, showURLs: bool = False
    ) -> t.List[t.Dict[str, t.Any]]:
        url = (
            self._base_url
            + "/"
            + str(indicator_id)
            + "?fields=descriptors&access_token="
            + self.api_token
        )

        if showURLs:
            print("url =", url)

        response = self.get_json_from_url(url)

        return response["descriptors"]["data"]



# Copyright (c) Meta Platforms, Inc. and affiliates.

from requests.adapters import HTTPAdapter


class TimeoutHTTPAdapter(HTTPAdapter):
    """
    Plug into requests to get a well-behaved session that does not wait for eternity.
    H/T: https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/#setting-default-timeouts
    """

    def __init__(self, *args, timeout=5, **kwargs):
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, *, timeout=None, **kwargs):
        if timeout is None:
            timeout = self.timeout
        return super().send(request, timeout=timeout, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.

import argparse
import binascii
import time
import pickle

import numpy
import faiss

from threatexchange.signal_type.pdq.pdq_utils import BITS_IN_PDQ

from threatexchange.signal_type.pdq.pdq_faiss_matcher import (
    PDQFlatHashIndex,
    PDQMultiHashIndex,
)

parser = argparse.ArgumentParser(
    description="Run basic benchmarks comparing PDQHashIndex implementations using faiss",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

parser.add_argument(
    "--faiss-threads",
    type=int,
    default=1,
    help="number of threads for faiss to use while searching",
)
parser.add_argument(
    "--dataset-size",
    type=int,
    default=250000,
    help="number of hashes to generate for the dataset to search against",
)
parser.add_argument(
    "--num-queries",
    type=int,
    default=1000,
    help="number of queries to generate for each search",
)
parser.add_argument(
    "--thresholds",
    type=int,
    default=[0, 15, 31, 47],
    choices=range(256),
    nargs="+",
    metavar="THRESHOLDS",
    help="PDQ similarity threshold values to benchmark with",
)
parser.add_argument("--seed", type=int, help="seed for random number generator")

args = parser.parse_args()

######
# Print Benchmark Settings
######

print("Benchmark: PDQ Faiss Matcher Comparison")
print("")
print("Options:")
for arg in vars(args):
    print("\t", arg, ": ", getattr(args, arg))
print("")

######
# Set up environment and helpers
######

faiss.omp_set_num_threads(args.faiss_threads)
seed = args.seed if args.seed else time.time_ns()
rng = numpy.random.default_rng(seed)
if args.seed is None:
    print("using random seed of ", seed)
    print("use --seed ", seed, " to rerun with same random values")
    print("")


def generate_random_hash():
    """
    returns a random 256 bit PDQ hash as a hexstring of 64 characters
    """
    hash_bytes = rng.bytes(BITS_IN_PDQ // 8)
    return binascii.hexlify(hash_bytes).decode()


def generate_random_distance_mask(hamming_distance):
    """
    returns a random numpy array of uint8s that can be used as bitwise mask
    to generate a hash with the given hamming distance
    """
    ones = numpy.ones(hamming_distance, dtype=numpy.uint8)
    bitmask = numpy.pad(
        ones, (0, BITS_IN_PDQ - hamming_distance), "constant", constant_values=0
    )
    return numpy.packbits(rng.permutation(bitmask))


def generate_random_hash_with_hamming_distance(original_hash, desired_hamming_distance):
    """
    returns a random 256 bit PDQ hash as a hexstring of 64 characters that is the given
    hamming distance from the provided original hash
    """
    original_hash_bytes = numpy.frombuffer(
        binascii.unhexlify(original_hash), dtype=numpy.uint8
    )
    mask = generate_random_distance_mask(desired_hamming_distance)
    new_hash_bytes = numpy.bitwise_xor(original_hash_bytes, mask).tobytes()
    return binascii.hexlify(new_hash_bytes).decode()


######
# Generate Random Dataset and Build Indexes
######

dataset = [generate_random_hash() for _ in range(args.dataset_size)]


custom_ids = [i + 100_000_000_000_000 for i in range(args.dataset_size)]

start_build_flat_hash_index = time.time()
flat_index = PDQFlatHashIndex()
flat_index.add(dataset, custom_ids=custom_ids)
serialized_flat_index = pickle.dumps(flat_index)
end_build_flat_hash_index = time.time()

start_build_multi_hash_index = time.time()
multi_index = PDQMultiHashIndex()
multi_index.add(dataset, custom_ids=custom_ids)
serialized_multi_index = pickle.dumps(multi_index)
end_build_multi_hash_index = time.time()

print("Building Stats:")

print(
    "\tPDQFlatHashIndex: time to build (s): ",
    end_build_flat_hash_index - start_build_flat_hash_index,
)
print(
    f"\tPDQFlatHashIndex: approximate size: {len(serialized_flat_index) // 1024:,d}KB"
)
print(
    "\tPDQMultiHashIndex: time to build (s): ",
    end_build_multi_hash_index - start_build_multi_hash_index,
)
print(
    f"\tPDQMultiHashIndex: approximate size: {len(serialized_multi_index) // 1024:,d}KB"
)
print("")

######
# Run benchmarks for each requested search threshold
######
for threshold in args.thresholds:
    print("Benchmarks for threshold: ", threshold)

    # Create queries with hamming distance of threshold compared to their search targets
    search_targets = rng.choice(dataset, size=args.num_queries)
    queries = [
        generate_random_hash_with_hamming_distance(target, threshold)
        for target in search_targets
    ]

    # Benchmark Searching Indexes
    start_flat_search = time.time()
    flat_results = flat_index.search(queries, threshold)
    end_flat_search = time.time()

    start_multi_search = time.time()
    multi_results = multi_index.search(queries, threshold)
    end_multi_search = time.time()

    def count_targets_found(targets, queries, results):
        """
        Checks that each element of the provided search results list contains
        the associated target for that query, warning if it does not.

        Returns the number of targets that were found in their corresponding
        results
        """
        found_targets = 0
        for target, query, result in zip(targets, queries, results):
            if target not in result:
                print(
                    "Query missed target in result: query=",
                    query,
                    "target=",
                    target,
                    "result=",
                    result,
                )
            else:
                found_targets += 1
        return found_targets

    flat_found_targets = count_targets_found(search_targets, queries, flat_results)
    multi_found_targets = count_targets_found(search_targets, queries, multi_results)

    print(
        "\tPDQFlatHashIndex - Total Time to search  (s): ",
        end_flat_search - start_flat_search,
    )
    print(
        "\tPDQMultiHashIndex - Total Time to search  (s): ",
        end_multi_search - start_multi_search,
    )
    print(
        "\tPDQFlatHashIndex - Percent of targets found: ",
        flat_found_targets / len(queries) * 100,
    )
    print(
        "\tPDQMultiHashIndex - Percent of targets found: ",
        multi_found_targets / len(queries) * 100,
    )

    print("")



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Wrapper around the pdf content type.
"""

from .content_base import ContentType


class URLContent(ContentType):
    """
    URLs are often used to point specific files and specific locations online.
    While a change in protocol/scheme is not necessarily significant to the location
    arrived at, the sub domains, domain and top level domains certainly are.
    Parameters are used differently depending on the service and so should be
    considered when hashing.

    So we first check that the protocol (including ':' but not '//' ) is removed
    (if not we remove it) and then simply MD5 hash remainder of the URL.

    URLs as per RFC 1738 [https://datatracker.ietf.org/doc/html/rfc1738] minus the
    scheme and following '//'

    Examples include:
    * www.facebook.com
    * drive.google.com/drive/u/0/folders/
    * discord.com/channels/32446207065914408/861844645281323795
    * twitter.com/userabc/status/1452649689363451231217/?lang=fr

    """



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Hash command to convert content into signatures.
"""

import argparse
import pathlib
import typing as t
import tempfile
from pathlib import Path

from threatexchange import common
from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli.exceptions import CommandError
from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.photo import PhotoContent
from threatexchange.content_type.file import FileContent
from threatexchange.content_type.content_base import RotationType

from threatexchange.signal_type.signal_base import FileHasher, SignalType
from threatexchange.cli import command_base
from threatexchange.cli.helpers import FlexFilesInputAction


class HashCommand(command_base.Command):
    """
    Take content and convert it into signatures (aka hashes)

    # Input
    You can pass in data to the command in a few different ways:
    (Note - you may not be able to hash text without an extension)
    ```
    # As an input file that contains one signal
    $ threatexchange hash photo my_photo.jpg

    # As stdin
    $ echo This is my cool text | threatexchange hash text -

    # Inline
    $ threatexchange hash text -- This is my cool text
    ```

    # Output
    <SignalType> <hash string>
    """

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        signal_types = [
            s for s in settings.get_all_signal_types() if issubclass(s, FileHasher)
        ]

        content_choices = sorted(s.get_name() for s in settings.get_all_content_types())
        signal_choices = sorted(
            s.get_name() for s in signal_types if issubclass(s, FileHasher)
        )

        ap.add_argument(
            "content_type",
            **common.argparse_choices_pre_type_kwargs(
                choices=content_choices,
                type=settings.get_content_type,
            ),
            help="what kind of content to hash",
        )

        ap.add_argument(
            "files",
            nargs=argparse.REMAINDER,
            action=FlexFilesInputAction,
            help="list of files, URLs, - for stdin, or -- to interpret remainder as a string",
        )

        ap.add_argument(
            "--signal-type",
            "-S",
            **common.argparse_choices_pre_type_kwargs(
                choices=signal_choices,
                type=settings.get_signal_type,
            ),
            help="only generate these signal types",
        )

        ap.add_argument(
            "--photo-preprocess",
            choices=["unletterbox", "rotations"],
            help=(
                "Apply one of the preprocessing steps to the image before hashing. "
                "'unletterbox' removes black borders, and 'rotations' generates all 8 "
                "simple rotations."
            ),
        )

        ap.add_argument(
            "--black-threshold",
            type=int,
            default=15,
            help=(
                "Set the black threshold for unletterboxing (default: 15)."
                "Only applies when 'unletterbox' is selected in --preprocess."
            ),
        )

        ap.add_argument(
            "--save-preprocess",
            action="store_true",
            help="save the preprocessed image data as new files",
        )

    def __init__(
        self,
        content_type: t.Type[ContentType],
        signal_type: t.Optional[t.Type[SignalType]],
        files: t.List[pathlib.Path],
        photo_preprocess: t.Optional[str] = None,
        black_threshold: int = 0,
        save_preprocess: bool = False,
    ) -> None:
        self.content_type = content_type
        self.signal_type = signal_type
        self.photo_preprocess = photo_preprocess
        self.black_threshold = black_threshold
        self.save_preprocess = save_preprocess
        self.files = files
        if self.photo_preprocess and not issubclass(self.content_type, PhotoContent):
            raise CommandError(
                "--photo-preprocess flag is only available for Photo content type", 2
            )
        if issubclass(self.content_type, FileContent):
            try:
                # Use the first file to determine content type
                self.content_type = FileContent.map_to_content_type(self.files[0])
            except ValueError as e:
                raise CommandError(f"{e}", returncode=2)

    def execute(self, settings: CLISettings) -> None:
        hashers = [
            s
            for s in settings.get_signal_types_for_content(self.content_type)
            if issubclass(s, FileHasher)
        ]
        if self.signal_type is not None:
            if self.signal_type not in hashers:
                raise CommandError.user(
                    f"{self.signal_type.get_name()} "
                    f"does not apply to {self.content_type.get_name()}"
                )

            hashers = [self.signal_type]  # type: ignore  # can't detect intersection types

        if not self.photo_preprocess:
            for file in self.files:
                for hasher in hashers:
                    hash_str = hasher.hash_from_file(file)
                    if hash_str:
                        print(hasher.get_name(), hash_str)
            return

        def pre_processed_files() -> (
            t.Iterator[t.Tuple[Path, bytes, t.Union[None, RotationType], str]]
        ):
            """
            Generator that yields preprocessed files and their metadata.
            Each item is a tuple of (file path, processed bytes, rotation name, image format).
            """
            for file in self.files:
                image_format = file.suffix.lower().lstrip(".")
                if self.photo_preprocess == "unletterbox":
                    processed_bytes = PhotoContent.unletterbox(
                        file, self.black_threshold
                    )
                    yield file, processed_bytes, None, image_format
                elif self.photo_preprocess == "rotations":
                    with open(file, "rb") as f:
                        image_bytes = f.read()
                    rotations = PhotoContent.all_simple_rotations(image_bytes)
                    for rotation_type, processed_bytes in rotations.items():
                        yield file, processed_bytes, rotation_type, image_format

        for (
            file,
            processed_bytes,
            rotation_type,
            image_format,
        ) in pre_processed_files():
            output_extension = f".{image_format.lower()}" if image_format else ".png"
            with tempfile.NamedTemporaryFile(
                delete=not self.save_preprocess, suffix=output_extension
            ) as temp_file:
                temp_file.write(processed_bytes)
                temp_file_path = Path(temp_file.name)
                for hasher in hashers:
                    hash_str = hasher.hash_from_file(temp_file_path)
                    if hash_str:
                        prefix = rotation_type.name if rotation_type else ""
                        print(f"{prefix} {hasher.get_name()} {hash_str}")
                if self.save_preprocess:
                    suffix = (
                        f"_{rotation_type.name}" if rotation_type else "_unletterboxed"
                    )
                    output_path = file.with_stem(f"{file.stem}{suffix}").with_suffix(
                        output_extension
                    )
                    temp_file_path.rename(output_path)
                    print(f"Processed image saved to: {output_path}")



# Copyright (c) Meta Platforms, Inc. and affiliates.

import io
import pdqhash
import pathlib
import numpy as np
from PIL import Image
import typing as t

PDQOutput = t.Tuple[
    str, int
]  # hexadecimal representation of the Hash vector and a numerical quality value


def pdq_from_file(path: pathlib.Path) -> PDQOutput:
    """
    Given a path to a file return the PDQ Hash string in hex.
    Current tested against: jpg
    """
    np_array = _convert_image_to_correct_array_dimension(Image.open(path))
    return _pdq_from_numpy_array(np_array)


def pdq_from_bytes(file_bytes: bytes) -> PDQOutput:
    """
    For the bytestream from an image file, compute PDQ Hash and quality.
    """
    np_array = _convert_image_to_correct_array_dimension(
        Image.open(io.BytesIO(file_bytes))
    )
    return _pdq_from_numpy_array(np_array)


def _pdq_from_numpy_array(array: np.ndarray) -> PDQOutput:
    hash_vector, quality = pdqhash.compute(array)

    bin_str = "".join([str(x) for x in hash_vector])

    # binary to hex using format string
    # '%0*' is for padding up to ceil(num_bits/4),
    # '%X' create a hex representation from the binary string's integer value
    hex_str = "%0*X" % ((len(bin_str) + 3) // 4, int(bin_str, 2))
    hex_str = hex_str.lower()

    return hex_str, quality


def _convert_image_to_correct_array_dimension(image: Image.Image) -> np.ndarray:
    """
    Handle possible image format conversion or
    """
    if image.mode == "LA" or image.mode == "I;16":
        # LA images (luminance with alpha) return 3 dimensional ndarray
        # which is incompatible with pdqhash
        image = image.convert("RGB")

    array = np.asarray(image)

    # Convert possible 2D array to 3D array
    # This is more efficient than converting to RGB for mode L images.
    if array.ndim == 2:
        array = np.concatenate([array[..., np.newaxis]] * 3, axis=2)

    return array



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Local storage and configuration for the CLI.

The CLI and Hasher-Matcher-Actioner are roughly parallel, but this isn't a
scalable service running on AWS. Instead, we have all of our state in
a file (likely ~/.threatexchange)
"""

from dataclasses import dataclass, field
import sys
import typing as t
import json
import pathlib
import logging

from dacite import WrongTypeError

from threatexchange.exchanges import collab_config
from threatexchange.exchanges.impl.stop_ncii_api import StopNCIICredentials
from threatexchange.exchanges.signal_exchange_api import (
    SignalExchangeAPI,
    TCollabConfig,
    TSignalExchangeAPICls,
)
from threatexchange.content_type import content_base
from threatexchange.exchanges import fetch_state
from threatexchange.exchanges.impl.static_sample import StaticSampleSignalExchangeAPI
from threatexchange.signal_type import signal_base
from threatexchange.interface_validation import FunctionalityMapping
from threatexchange.cli.cli_state import CliSimpleState, CliIndexStore
from threatexchange.storage import interfaces as iface
from threatexchange.utils import dataclass_json

CONFIG_FILENAME = "config.json"


@dataclass
class CLiConfig:
    """A place to store misc configuration for the CLI"""

    fb_threatexchange_api_token: t.Optional[str] = None
    ncmec_credentials: t.Optional[t.Tuple[str, str]] = None
    tat_credentials: t.Optional[t.Tuple[str, str]] = None
    stop_ncii_keys: t.Optional[StopNCIICredentials] = None
    extensions: t.Set[str] = field(default_factory=set)
    # Every item needs a default for backwards compatibility


class CliState(collab_config.CollaborationConfigStoreBase):
    """
    A wrapper around stateful information stored for the CLI.

    Everything is just in a single directory (usually ~/.threatexchange).
    """

    def __init__(
        self,
        fetch_types: t.List[t.Union[SignalExchangeAPI, t.Type[SignalExchangeAPI]]],
        dir: pathlib.Path,
    ):
        self._dir = dir.expanduser()

        self._name_to_ctype = {ft.get_name(): ft.get_config_cls() for ft in fetch_types}

        self._cache: t.Optional[t.Dict[str, collab_config.CollaborationConfigBase]] = (
            None
        )

        self._init_folders_if_needed()

    def _init_folders_if_needed(self):
        for d in (self.collab_dir, self.index_dir, self.fetched_state_dir):
            if not d.is_dir():
                d.mkdir(parents=True)
        cfg = self.config_file
        if not cfg.is_file():
            cfg.write_text("{}")

    @property
    def collab_dir(self) -> pathlib.Path:
        return self._dir / "collab_configs/"

    @property
    def fetched_state_dir(self) -> pathlib.Path:
        return self._dir / "fetched"

    @property
    def index_dir(self) -> pathlib.Path:
        return self._dir / "index/"

    @property
    def config_file(self) -> pathlib.Path:
        return self._dir / "config.json"

    def path_for_collab_config(
        self, config: collab_config.CollaborationConfigBase
    ) -> pathlib.Path:
        return self.collab_dir / f"{config.name}.json"

    def get_persistent_config(self) -> CLiConfig:
        return dataclass_json.dataclass_load_file(
            self.config_file, CLiConfig, default=CLiConfig()
        )

    def update_persistent_config(self, config: CLiConfig):
        dataclass_json.dataclass_dump_file(self.config_file, config)

    def dir_for_fetched_state(
        self,
        api: t.Type[SignalExchangeAPI],
    ) -> pathlib.Path:
        return self.fetched_state_dir / f"{api.get_name()}/"

    def get_collab_names_without_loading(self) -> t.List[str]:
        if self._cache is not None:
            return list(self._cache)
        return [str(p) for p in self.collab_dir.glob("*.json")]

    def get_all_collabs(self) -> t.List[collab_config.CollaborationConfigBase]:
        """
        Get all CollaborationConfigs, already resolved to the correct type
        """
        if self._cache is None:
            collab_dir = self.collab_dir

            ret = []
            for f in collab_dir.glob("*.json"):
                if not f.is_file():
                    logging.warning("Ignoring strange file in collab dir: %s", f)
                    continue
                with f.open() as fp:
                    try:
                        content = json.load(fp)
                    except json.JSONDecodeError:
                        logging.exception("Failed to parse collab config: %s", f)
                        continue
                    ctype = None
                    if isinstance(content, dict):
                        ctype = self._name_to_ctype.get(content.get("api"))  # type: ignore
                    if ctype is None:
                        logging.warning("Ignoring collab config of unknown type: %s", f)
                        continue
                try:
                    config = dataclass_json.dataclass_load_dict(content, ctype)
                    ret.append(config)
                except WrongTypeError:
                    logging.exception("Failed to parse collab config: %s", f)
            self._cache = {c.name: c for c in ret}
        return list(self._cache.values())

    def update_collab(self, collab: collab_config.CollaborationConfigBase) -> None:
        """Create or update a collaboration"""
        assert collab.api, "didn't set API?"
        path = self.path_for_collab_config(collab)
        dataclass_json.dataclass_dump_file(path, collab)

    def delete_collab(self, collab: collab_config.CollaborationConfigBase) -> None:
        """Delete a collaboration"""
        self.path_for_collab_config(collab).unlink(missing_ok=True)


@dataclass
class _SignalExchangeAccessor:
    """Convenience wrapper for operations on the SignalExchangeAPI"""

    _parent: "CLISettings"

    def get_all(self) -> t.ValuesView[TSignalExchangeAPICls]:
        return self._parent._mapping.exchange.api_by_name.values()

    def get_instance_for_collab(self, collab: TCollabConfig) -> SignalExchangeAPI[
        TCollabConfig,
        fetch_state.FetchCheckpointBase,
        fetch_state.FetchedSignalMetadata,
        t.Any,
        t.Any,
    ]:
        api_cls = self._parent._mapping.exchange.api_by_name[collab.api]
        return api_cls.for_collab(collab)

    def __iter__(self) -> t.Iterator[TSignalExchangeAPICls]:
        yield from self.get_all()


@dataclass
class _FetchStoreAccessor:
    """Convenience wrapper for operations on the state"""

    _parent: "CLISettings"

    def empty(self) -> bool:
        """Return the collabs with stored state"""
        collabs = self._parent.get_all_collabs()
        return not any(
            collab for collab in collabs if self.get_for_collab(collab).exists(collab)
        )

    def get_for_api(self, api: t.Type[SignalExchangeAPI]) -> CliSimpleState:
        return CliSimpleState(api, self._parent._state.dir_for_fetched_state(api))

    def get_for_collab(
        self, collab: collab_config.CollaborationConfigBase
    ) -> CliSimpleState:
        return self.get_for_api(self._parent._mapping.exchange.api_by_name[collab.api])


class CLISettings:
    """
    A God object for all miscellanious persisted state to make the CLI work
    """

    def __init__(
        self,
        mapping: FunctionalityMapping,
        cli_state: CliState,
    ) -> None:
        self._mapping = mapping
        self._state = cli_state
        self._sample_message_printed = False
        self._config: t.Optional[CLiConfig] = None
        self.index = CliIndexStore(cli_state.index_dir)
        self.fetched_state = _FetchStoreAccessor(self)
        self.apis = _SignalExchangeAccessor(self)
        self.new_iface = CLICompatibilityStorage(self)

    def get_persistent_config(self) -> CLiConfig:
        if self._config is None:
            self._config = self._state.get_persistent_config()
        return self._config

    def set_persistent_config(self, config: CLiConfig) -> None:
        self._state.update_persistent_config(config)
        self._config = config

    def get_all_content_types(self) -> t.List[t.Type[content_base.ContentType]]:
        return list(self._mapping.signal_and_content.content_by_name.values())

    def get_content_type(self, name: str) -> t.Type[content_base.ContentType]:
        return self._mapping.signal_and_content.content_by_name[name]

    def get_all_signal_types(self) -> t.List[t.Type[signal_base.SignalType]]:
        return list(self._mapping.signal_and_content.signal_type_by_name.values())

    def get_signal_type(self, name: str) -> t.Type[signal_base.SignalType]:
        return self._mapping.signal_and_content.signal_type_by_name[name]

    def get_signal_types_for_content(
        self, content_type: t.Type[content_base.ContentType]
    ) -> t.List[t.Type[signal_base.SignalType]]:
        return self._mapping.signal_and_content.signal_type_by_content[content_type]

    @property
    def in_demo_mode(self) -> bool:
        """Has no live collabs"""
        return not self._state.get_all_collabs()

    def get_all_collabs(
        self, *, default_to_sample: bool = True
    ) -> t.List[collab_config.CollaborationConfigBase]:
        if self.in_demo_mode and default_to_sample:
            return [self._get_sample_collab()]
        # Should this check whether the APIs are all valid?
        return self._state.get_all_collabs()

    def get_collab(
        self,
        name: str,
    ) -> t.Optional[collab_config.CollaborationConfigBase]:
        return self._state.get_collab(name)

    def _get_sample_collab(self) -> collab_config.CollaborationConfigBase:
        if not self._sample_message_printed:
            print(
                (
                    "Looks like you haven't set up a collaboration config, "
                    "so using the sample one against sample data"
                ),
                file=sys.stderr,
            )
            self._sample_message_printed = True
        return collab_config.CollaborationConfigBase(
            "Sample Signals", StaticSampleSignalExchangeAPI.get_name(), enabled=True
        )

    def get_collabs_for_api(
        self, api: SignalExchangeAPI
    ) -> t.List[collab_config.CollaborationConfigBase]:
        api_name = api.get_name()
        return [c for c in self.get_all_collabs() if c.api == api_name]


# TODO - eventually to unified store
class CLICompatibilityStorage(iface.ISignalTypeConfigStore):
    """
    Translate the new-style interface to the previous version.

    The goal is to eventually allow the CLI to use the storage interface
    for the operations that make sense for it, to allow using the CLI to
    test new storage interfaces.

    How to make a little progress on the migration:
      1. Implement a little more of the iface classes
      2. Change the equivalent interface on the old storage from
         method_name -> _method_name (new interface calls _method_name())
      3. Fix all the errors that appear, by calling
         old_iface.new_iface.new_method()
      4. Move the body of the implementation into this class

    How to make faster progress on the migration:
      1. Refactor the base command to take the new iface instead of the old one
      2. Use .old_iface() as needed

    You may eventually run into something that doesn't translate at all, which
    might require changing some of the underlying commands. Here are some
    tricky ones to watch out for:
      1. Some persistent data doesn't have an equivalent on the interface -
         e.g. extensions. That's fine, those will always be CLI specific, and
         the eventual model is probably one where the CLI config "contains" the
         interface elements (allowing you to swap storage classes), rather than
         being one.
    """

    def __init__(self, old_iface: CLISettings) -> None:
        self.old_iface = old_iface

    def get_signal_type_configs(self) -> t.Mapping[str, iface.SignalTypeConfig]:
        """Return all installed signal types."""
        # TODO replace with body of old_iface impl after refactor
        return {
            st.get_name(): iface.SignalTypeConfig(
                enabled_ratio=1.0,
                signal_type=st,
            )
            for st in self.old_iface.get_all_signal_types()
        }

    def _create_or_update_signal_type_override(
        self, signal_type: str, enabled_ratio: float
    ) -> None:
        raise NotImplementedError("Not yet supported")



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
A way to connect all the key interfaces of the library.

This is entirely optional, and you can always use the classes directly
without any trouble, but you may find some of the validation methods
useful, especially if you are relying on the various .get_name() functions
as storage keys, since there are helpers here for asserting they are
all unique.

Since matching may need to look up context from the original fetching,
this is all the parts needed to get you there.

You don't need to do it this way, but it does make it clear which
fetching, matching, and other capabilities you are supporting.
"""

import typing as t
from dataclasses import dataclass

from threatexchange.content_type.content_base import ContentType
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.exchanges.signal_exchange_api import (
    TSignalExchangeAPICls,
)
from threatexchange.exchanges.collab_config import CollaborationConfigStoreBase


class SignalTypeMapping:
    def __init__(
        self,
        content_types: t.List[t.Type[ContentType]],
        signal_types: t.List[t.Type[SignalType]],
    ):
        _validate_content_and_signal(content_types, signal_types)

        self.content_by_name = {c.get_name(): c for c in content_types}
        self.signal_type_by_name = {s.get_name(): s for s in signal_types}
        self.signal_type_by_content: t.Dict[
            t.Type[ContentType], t.List[t.Type[SignalType]]
        ] = {c: [] for c in content_types}
        for signal_type in signal_types:
            for content_type in signal_type.get_content_types():
                self.signal_type_by_content[content_type].append(signal_type)

    def get_supported_signal_types_for_content(
        self, content: t.Type[ContentType]
    ) -> t.List[t.Type[SignalType]]:
        # TODO - consider storing by tuple instead of defending against accidental
        #        mutation by callers
        return list(self.signal_type_by_content.get(content, ()))


class SignalExchangeAPIMapping:
    def __init__(self, apis: t.Sequence[TSignalExchangeAPICls]) -> None:
        _validate_signal_apis(apis)
        self.api_by_name = {f.get_name(): f for f in apis}


@dataclass
class FunctionalityMapping:
    """
    All of key fetch, hash, match interfaces combined.

    Since matching may need to look up context from the original fetching,
    this container provides all the interfaces needed to get you there.
    """

    signal_and_content: SignalTypeMapping
    exchange: SignalExchangeAPIMapping
    collabs: CollaborationConfigStoreBase


def _validate_signal_apis(apis: t.Iterable[TSignalExchangeAPICls]):
    names = set()
    for a in apis:
        name = a.get_name()
        assert name not in names, f"Duplicate name in {a.__name__}s: '{name}'"
        names.add(name)


def _validate_content_types(content_types: t.List[t.Type[ContentType]]) -> None:
    names = set()
    for c in content_types:
        name = c.get_name()
        assert name not in names, f"Duplicate name in {ContentType.__name__}s: '{name}'"
        names.add(name)


def _validate_signal_types(signal_types: t.List[t.Type[SignalType]]):
    names = set()
    for s in signal_types:
        name = s.get_name()
        assert name not in names, f"Duplicate name in {SignalType.__name__}s: '{name}'"
        names.add(name)


def _validate_content_and_signal(
    content_types: t.List[t.Type[ContentType]],
    signal_types: t.List[t.Type[SignalType]],
) -> None:
    _validate_content_types(content_types)
    _validate_signal_types(signal_types)
    supported_ct = set(content_types)
    for signal_type in signal_types:
        supported = any(c in supported_ct for c in signal_type.get_content_types())
        assert supported, f"No signal types for content type: {signal_type.get_name()}"



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Abstraction for different content types.

This records all the valid signal types for a piece of content.
"""

from enum import Enum
import typing as t

from threatexchange import common


class ContentType:
    @classmethod
    def get_name(cls) -> str:
        """The human-friendly display name"""
        return common.class_name_to_human_name(cls.__name__, "Content")

    @classmethod
    def extract_additional_content(
        cls, content_arg: str
    ) -> t.List[t.Tuple[t.Type["ContentType"], str]]:
        """
        Post-process/download content to find additional components

        Examples might be:
        * URL => download page content and return photo/video/text/URL snippets
        * File => Identify content based on file type and return appropriate snippets
        * Photo => run OCR and extract text
        * Video => break out photo thumbnail, close caption text, audio
        """
        return []


class RotationType(Enum):
    """
    Enum for 8 simple rotations of an image.
    Used to store all generated rotations of an image,
    whose algorithms don't have a native way to generate rotations during hashing.
    """

    ORIGINAL = "original"  # No rotation; the object is in its original orientation
    ROTATE90 = "rotate90"  # Rotates the object 90 degrees
    ROTATE180 = "rotate180"  # Rotates the object 180 degrees (half-turn)
    ROTATE270 = "rotate270"  # Rotates the object 270 degrees
    FLIPX = "flipx"  # Flip the object horizontally along the X-axis
    FLIPY = "flipy"  # Flip the object horizontally along the Y-axis
    FLIPPLUS1 = "flipplus1"  # Diagonal flip along the line y = x
    FLIPMINUS1 = "flipminus1"  # Diagonal flip along the line y = -x



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Wrapper around the Photo PDQ plus OCR signal type.
Requires the installations of Optical character recognition (OCR) tools
"""

import typing as t
import pathlib

from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.photo import PhotoContent

from threatexchange.signal_type.pdq.signal import PdqSignal
from threatexchange.signal_type.raw_text import RawTextSignal
from threatexchange.signal_type import signal_base
from threatexchange.exchanges.impl.fb_threatexchange_signal import (
    HasFbThreatExchangeIndicatorType,
)

from threatexchange.signal_type.pdq.pdq_hasher import pdq_from_file
from threatexchange.extensions.pdq_ocr.ocr_utils import text_from_image_file


class PdqOcrSignal(
    signal_base.SimpleSignalType,
    signal_base.FileHasher,
    HasFbThreatExchangeIndicatorType,
):
    """
    PDQ is an open source photo similarity algorithm. See 'pdq.py'
    This signal type combines pdq hashes with a text string found using
    optical character recognition (OCR) on the image hashed. This adds an
    additional 'distance' using levenshtein string comparison.

    Accuracy of matches will depend on quality of OCR implementations and selected thresholds.
    (ocr_utils uses the open source pytesseract wrapper of tesseract)
    """

    INDICATOR_TYPE = "HASH_PDQ_OCR"

    # This may need to be updated (TODO make more configurable)
    # Hashes of distance less than or equal to this threshold are considered a 'match'
    PDQ_PLUS_OCR_CONFIDENT_MATCH_THRESHOLD = 31
    # Match considered if 90% of the strings match
    LEVENSHTEIN_DISTANCE_PERCENT_THRESHOLD = 90

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [PhotoContent]

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        pdq_hash, quality = pdq_from_file(file)
        ocr_text = text_from_image_file(file)

        return f"{pdq_hash},{ocr_text}"

    @classmethod
    def compare_hash(
        cls,
        hash1: str,
        hash2: str,
        pdq_dist_threshold: int = PDQ_PLUS_OCR_CONFIDENT_MATCH_THRESHOLD,
    ) -> signal_base.SignalComparisonResult:
        assert 0 <= pdq_dist_threshold <= 256
        pdq_hash_1, _, ocr_text_1 = hash1.partition(",")
        pdq_hash_2, _, ocr_text_2 = hash2.partition(",")
        assert all(
            (pdq_hash_1, ocr_text_1, pdq_hash_2, ocr_text_2)
        ), "malformed pdq_ocr hash"

        pdq_result = PdqSignal.compare_hash(pdq_hash_1, pdq_hash_2, pdq_dist_threshold)
        if not pdq_result.match:
            return pdq_result
        text_result = RawTextSignal.matches_str(
            ocr_text_1, ocr_text_2, cls.LEVENSHTEIN_DISTANCE_PERCENT_THRESHOLD
        )
        return signal_base.SignalComparisonResult(
            text_result.match, pdq_result.distance
        )

    @staticmethod
    def get_examples() -> t.List[str]:
        return [
            "a72dd3eadec2800ba74b59a9532d0b22011fd9e0daa1da3f576e602db999a754,This is a sample text string",
            "b2539a60de78841da72fcdeb5da21bf00185632ddaadb23d174ea1885999cb65,This is a sample text string",
        ]



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Helper to allow you to hint at the ThreatExchange type string on SignalTypes
"""

import typing as t


class HasFbThreatExchangeIndicatorType:
    """
    A mixin to hint to fb_threatexchange_api how to handle this SignalType

    For example, PDQSignalType, aka "pdq" in ThreatExchange is "HASH_PDQ"

    We could hardcode all of these into the fetch() method, but that doesn't
    play nicely with expansions. Our solution:

    ```
    class PDQSignalType(SignalType, HasFbThreatExchangeIndicatorType):
        INDICATOR_TYPE = "HASH_PDQ"
        ...
    ```

    Some types might have multiple representations in ThreatExchange, like URL:

    ```
    INDICATOR_TYPE = ("URI", "RAW_URI", "UNCLICKABLE_URL")
    ```

    One of the strengths of ThreatExchange is its flexibility in conventions.
    As long as it takes less than 1MB of storage, we can serialize any data using
    the DEBUG_STRING indicator type. For that, we need to also use the tags feature.
    One convention is to use a `signal_type:` prefix.

    ```
    INDICATOR_TYPE = {"DEBUG_STRING": "signal_type:my_prototype_signal"}
    ```
    Later one it has a real type, you can update to
    ```
    INDICATOR_TYPE = {
        "MY_PROTOTYPE_SIGNAL": None,
        "DEBUG_STRING": "signal_type:my_prototype_signal",
    }
    ```
    """

    INDICATOR_TYPE: t.ClassVar[
        t.Union[str, t.Set[str], t.Dict[str, t.Optional[str]]]
    ] = set()

    @classmethod
    def normalize_fb_threatexchange_indicator(
        cls, tx_type: str, tx_indicator: str, tx_tag: t.Optional[str]
    ) -> str:
        """
        Cleanup signals that might change format depending on type.

        Example:
        RAW_URI: https://www.facebook.com
        UNCLICKABLE_URL: [h]ttps://www.facebook.com

        Post normalized: https://www.facebook.com
        """
        return tx_indicator



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""Simple implementation for the Tech Against Terrorism Hash List REST API"""

from dataclasses import dataclass
import datetime
from enum import Enum
import requests
import logging
import typing as t
from contextlib import contextmanager
import requests
from requests.packages.urllib3.util.retry import Retry
from threatexchange.exchanges.clients.utils.common import TimeoutHTTPAdapter


class TATIdeology(Enum):
    islamist = "islamist"
    far_right = "far-right"
    _all = "all"


@dataclass
class TATAPIErrorResponse:
    error: str


@dataclass
class TATHashListEntry:
    hash_digest: str
    algorithm: str
    ideology: TATIdeology
    file_type: str
    deleted: bool
    updated_on: float
    created_on: datetime.datetime
    id: int


@dataclass
class TATHashListResponse:
    count: int
    next: t.Optional[str]
    previous: t.Optional[str]
    checkpoint: str
    results: t.List[TATHashListEntry]


@dataclass
class TATUser:
    user: t.Dict[str, t.Any]
    token: str


class TATEndpoint(Enum):
    authenticate = "token-auth/tcap/"
    hash_list = "api/hash-list/v2/all"


class TATHashListAPI:
    """
    A wrapper around the Tech Against Terrorism Hash List API

    The verification and collection of terrorist content are conducted by TAT OSINT Analysts and automated processes.
    Subsequently, the content is classified and hashed by the TAT Archive hashing and classification services.

    The list is udpated in real-time as content is classified or captured.

    For more information on our collection process please visit: https://terrorismanalytics.org/about/how-it-works
    For our Hash List documentation: https://terrorismanalytics.org/docs/hash-list-v1
    """

    BASE_URL: t.ClassVar[str] = "https://app.terrorismanalytics.org/"

    def __init__(self, username: str, password: str) -> None:
        self.username = username
        self.password = password

    @contextmanager
    def _get_session(self, auth_token: t.Optional[str] = None):
        session = requests.Session()
        session.mount(
            self.BASE_URL,
            adapter=TimeoutHTTPAdapter(
                timeout=60,
                max_retries=Retry(
                    total=4,
                    status_forcelist=[429, 500, 502, 503, 504],
                    allowed_methods=["HEAD", "GET", "OPTIONS"],
                    backoff_factor=0.2,  # ~1.5 seconds of retries
                ),
            ),
        )

        if auth_token:
            session.headers.update({"Authorization": f"Bearer {auth_token}"})

        try:
            yield session
        finally:
            session.close()

    def _get(
        self,
        full_url: t.Optional[str] = None,
        auth_token: t.Optional[str] = None,
        **params,
    ) -> t.Any:
        """
        Perform an HTTP GET request, and return the JSON response payload.

        Same timeouts and retry strategy as `_get_session` above.
        """

        with self._get_session(auth_token) as session:
            response = session.get(url=full_url, params=params)
            response.raise_for_status()
            return response.json()

    def _post(self, endpoint: TATEndpoint, data=None) -> t.Any:
        """
        Perform an HTTP POST request, and return the JSON response payload.

        Same timeouts and retry strategy as `_get_session` above.
        """
        with self._get_session() as session:
            url = self.BASE_URL + endpoint.value
            response = session.post(url=url, data=data)
            response.raise_for_status()
            return response.json()

    def get_auth_token(self, username: str, password: str) -> t.Optional[str]:
        """
        Authenticate with TCAP services and obtain a JWT token
        """

        logging.info("Authenticating with TCAP: %s", username)

        auth_response = self._post(
            TATEndpoint.authenticate,
            data={"username": username, "password": password, "resend": False},
        )
        return auth_response.get("token")

    def fetch_hashes(
        self,
        next_page: t.Optional[str] = None,
        order: str = "asc",
        after: str = "",
    ) -> TATHashListResponse:
        """
        Get the Hash List JSON file presigned URL ( 5 Minute expiry ) and metadata
        """

        params: t.Dict[str, t.Any] = {
            "order": order,
            "after": after,
        }

        try:
            token = self.get_auth_token(self.username, self.password)
            endpoint = f"{TATEndpoint.hash_list.value}"
            url = f"{self.BASE_URL}{endpoint}?{next_page}"

            logging.info("Fetching TAT hash list")

            results = self._get(full_url=url, auth_token=token, **params)

            return TATHashListResponse(
                count=results["count"],
                next=results["next"],
                previous=results["previous"],
                checkpoint=results["checkpoint"],
                results=[TATHashListEntry(**entry) for entry in results["results"]],
            )

        except Exception as exception:
            logging.error("Failed to get hash list: %s", exception)
            raise

    def fetch_hashes_iter(
        self, checkpoint: str = ""
    ) -> t.Iterator[TATHashListResponse]:
        """
        A wrapper to continously fetch the hash list in a paginated manner.
        Each page has 3 elements we're interested in here:

        - next (str | None): A URL with limit and offset query params automatically applied if there are more results.
        - checkpoint (str): A timestamp and id of the last record in the current page eg: 1704085200,124
        - results (List[TATHashListEntry]): A list of hash list entries

        if next in null we have no more hashes left to fetch meaning we break out of the loop
        """

        has_more = True
        next_page = None

        while has_more:
            response = self.fetch_hashes(next_page=next_page, after=checkpoint)
            has_more = bool(response.next)

            if has_more:
                next_page = response.next

            yield response



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

import typing as t

from PIL import Image


def is_pixel_black(pixel: tuple, black_threshold: int):
    """
    Check if each color channel in the pixel is below the threshold
    """
    r, g, b = pixel
    return r <= black_threshold and g <= black_threshold and b <= black_threshold


def detect_top_border(image: Image.Image, black_threshold: int = 0) -> int:
    """
    Detect the top black border by counting rows with only black pixels.
    Checks each RGB channel of each pixel in each row.
    Returns the first row that is not all black from the top.
    """
    width, height = image.size
    for y in range(height):
        row_pixels = list(image.crop((0, y, width, y + 1)).get_flattened_data())
        if all(
            is_pixel_black(t.cast(t.Tuple[t.Any, ...], pixel), black_threshold)
            for pixel in row_pixels
        ):
            continue
        return y
    return height


def detect_bottom_border(image: Image.Image, black_threshold: int = 0) -> int:
    """
    Detect the bottom black border by counting rows with only black pixels from the bottom up.
    Checks each RGB channel of each pixel in each row.
    Returns the first row that is not all black from the bottom.
    """
    width, height = image.size
    for y in range(height - 1, -1, -1):
        row_pixels = list(image.crop((0, y, width, y + 1)).get_flattened_data())
        if all(
            is_pixel_black(t.cast(t.Tuple[t.Any, ...], pixel), black_threshold)
            for pixel in row_pixels
        ):
            continue
        return height - y - 1
    return height


def detect_left_border(image: Image.Image, black_threshold: int = 0) -> int:
    """
    Detect the left black border by counting columns with only black pixels.
    Checks each RGB channel of each pixel in each column.
    Returns the first column from the left that is not all black.
    """
    width, height = image.size
    for x in range(width):
        col_pixels = list(image.crop((x, 0, x + 1, height)).get_flattened_data())
        if all(
            is_pixel_black(t.cast(t.Tuple[t.Any, ...], pixel), black_threshold)
            for pixel in col_pixels
        ):
            continue
        return x
    return width


def detect_right_border(image: Image.Image, black_threshold: int = 0) -> int:
    """
    Detect the right black border by counting columns with only black pixels from the right.
    Checks each RGB channel of each pixel in each column.
    Returns the first column from the right that is not all black.
    """
    width, height = image.size
    for x in range(width - 1, -1, -1):
        col_pixels = list(image.crop((x, 0, x + 1, height)).get_flattened_data())
        if all(
            is_pixel_black(t.cast(t.Tuple[t.Any, ...], pixel), black_threshold)
            for pixel in col_pixels
        ):
            continue
        return width - x - 1
    return width



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Wrapper around the URL signal type.
"""

import typing as t

from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.url import URLContent
from threatexchange.signal_type import signal_base
from threatexchange.exchanges.impl.fb_threatexchange_signal import (
    HasFbThreatExchangeIndicatorType,
)


class URLSignal(
    signal_base.SimpleSignalType,
    signal_base.MatchesStr,
    HasFbThreatExchangeIndicatorType,
):
    """
    Wrapper around URL links, such as https://github.com/
    """

    INDICATOR_TYPE = {"URI", "RAW_URI"}

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [URLContent]

    @classmethod
    def get_index_cls(cls) -> t.Type[signal_base.TrivialSignalTypeIndex]:
        # TODO - There are a lot more considerations than this
        #        as well as normalizations that could be applied
        return signal_base.TrivialSignalTypeIndex

    @classmethod
    def matches_str(
        cls, signal: str, haystack: str
    ) -> signal_base.SignalComparisonResult:
        # TODO - normalization
        return signal_base.SignalComparisonResult.from_bool_only(signal == haystack)

    @staticmethod
    def get_examples() -> t.List[str]:
        return ["https://developers.facebook.com/docs/threat-exchange/reference/apis/"]

    @classmethod
    def normalize_fb_threatexchange_indicator(
        cls, tx_type: str, tx_indicator: str, tx_tag: t.Optional[str]
    ) -> str:
        if tx_type == "UNCLICKABLE_URL" and tx_indicator.startswith("[h]"):
            return f"h{tx_indicator[3:]}"
        return super().normalize_fb_threatexchange_indicator(
            tx_type, tx_indicator, tx_tag
        )
